"""Operator graph optimizers."""

import logging
import warnings
from collections import defaultdict, namedtuple
from itertools import zip_longest

import numpy as np

from nengo.builder.neurons import SimNeurons
from nengo.builder.operator import (
    BsrDotInc,
    Copy,
    DotInc,
    ElementwiseInc,
    Reset,
    reshape_dot,
)
from nengo.builder.signal import Signal
from nengo.rc import rc
from nengo.utils.graphs import BidirectionalDAG, transitive_closure
from nengo.utils.stdlib import Timer, WeakKeyDefaultDict, WeakSet

logger = logging.getLogger(__name__)


def optimize(model, dg):
    """Optimizes the operator graph by merging operators.

    This reduces the number of iterators to iterate over in slow Python code
    (as opposed to fast C code). The resulting merged operators will also
    operate on larger chunks of sequential memory, making better use of CPU
    caching and prefetching.

    The optimization algorithm has worst case complexity :math:`O(n^2 + e)`,
    where :math:`n` is the number of operators and :math:`e` is the number
    of edges in the dependency graph. In practice the run time will be much
    better because not all :math:`n^2` pairwise combinations of operators
    will be evaluated. A grouping depending on the operator type and view
    bases is done with dictionaries. This grouping can be done in amortized
    linear time and reduces the actual worst-case runtime of the optimization
    algorithm to :math:`O(gm^2 + e)`, where :math:`g` is the number of groups
    and :math:`m` is the number of elements in a group. Moreover, information
    about memory alignment will be used to cut the inner loop short in
    many cases and gives a runtime much closer to linear in most cases.

    Note that this function modifies both ``model`` and ``dg``.

    Parameters
    ----------
    model : `nengo.builder.Model`
        Builder output to optimize.
    dg : dict
        Dict of the form ``{a: {b, c}}`` where ``b`` and ``c`` depend on ``a``,
        specifying the operator dependency graph of the model.
    """

    logger.debug("Optimizing model...")

    # We try first to merge operators with views only as these have a fixed
    # order for the memory alignment whereas operators without views could
    # be merged in a random order. Merging the views of operators will
    # propagate requirements in the memory ordering via the other
    # associated signals of the operator to other operators.

    # Once no more operators with views can be merged, we try to merge
    # operators without views and then try again merging views (because
    # each operator merge might generate new views).

    single_pass = OpMergePass(dg)

    n_initial_ops = len(dg)
    cum_duration = 0.0
    before, after = None, None
    only_merge_ops_with_view = True
    while only_merge_ops_with_view or after < before:
        only_merge_ops_with_view = before is None or before != after
        before = len(single_pass.dg.forward)

        with Timer() as t:
            single_pass(only_merge_ops_with_view)

        after = len(single_pass.dg.forward)
        logger.debug(
            "[%s]: Reduced %i to %i operators in %fs.",
            "views" if only_merge_ops_with_view else "non-views",
            before,
            after,
            t.duration,
        )

        # Prevent optimizer from running too long if we get up diminishing
        # returns.
        # Note that we don't break if there was no reduction at all because
        # in that case we want to toggle only_merge_ops_with_view which might
        # still yield some significant reduction.
        cum_duration += t.duration
        mean_reduction_rate = float(n_initial_ops - after) / cum_duration
        last_reduction_rate = float(before - after) / t.duration
        threshold = 0.01
        scaled_rate = threshold * mean_reduction_rate
        if 0.0 < last_reduction_rate < scaled_rate:  # pragma: no cover
            logger.debug(
                "Operator reduction rate fell below threshold of %.3f. "
                "Stopping optimizer.",
                threshold,
            )
            break

    # Update model signals
    for sigdict in model.sig.values():
        for name in sigdict:
            while sigdict[name] in single_pass.sig_replacements:
                sigdict[name] = single_pass.sig_replacements[sigdict[name]]

    # Reinitialize the model's operator list
    del model.operators[:]
    for op in dg:
        model.add_op(op)


class OpMergePass:
    """Manages a single optimization pass."""

    def __init__(self, dg):
        self.dg = BidirectionalDAG(dg)
        self.might_merge = set(dg)
        self.sig_replacements = {}

        self.sig2ops = WeakKeyDefaultDict(WeakSet)
        self.base2views = WeakKeyDefaultDict(WeakSet)
        for op in self.dg.forward:
            for s in op.all_signals:
                self.sig2ops[s].add(op)
                self.base2views[s.base].add(s)

        # These variables will be initialized and used on each pass
        self.dependents = None
        self.only_merge_ops_with_view = None

        self.merged = set()
        self.merged_dependents = set()
        self.opinfo = OpInfo()

    def __call__(self, only_merge_ops_with_view):
        """Perform a single optimization pass.

        Parameters
        ----------
        only_merge_ops_with_view : bool
            Limits operator merges to operators with views.
        """

        # --- Initialize pass state
        self.dependents = transitive_closure(self.dg.forward)
        self.only_merge_ops_with_view = only_merge_ops_with_view
        self.merged.clear()
        self.merged_dependents.clear()
        self.opinfo.clear()

        # --- Do an optimization pass
        self.perform_merges()

    def perform_merges(self):
        """Go through all operators and merge them where possible.

        Parameters
        ----------
        only_merge_ops_with_view : bool
            Limit merges to operators with views.
        """

        # We go through the ops grouped by type as only ops with the same
        # type can be merged.
        by_type = groupby(self.might_merge, type)

        # Note that we will stop once we merge any operator, so merges are
        # performed on at most one type of operator per pass.
        # The dependency graph and other information will be updated
        # before merging other operator types.

        # We go through ops in a heuristic order to reduce runtime
        firstops = [ElementwiseInc, Copy, DotInc, SimNeurons]
        sortedops = firstops + [op for op in by_type if op not in firstops]
        for optype in sortedops:

            if OpMerger.is_type_mergeable(optype):
                self.perform_merges_for_subset(by_type[optype])

            # If we're not only merging views, the memory layout changes
            # and non-views are turned into views. In that case we need
            # to update the signals the operators are referring to before
            # trying to merge a different type of operators. Thus, we break
            # the loop here.
            if not self.only_merge_ops_with_view and len(self.merged) > 0:
                break

    def perform_merges_for_subset(self, subset):
        """Performs operator merges for a subset of operators.

        Parameters
        ----------
        subset : list
            Subset of operators.
        """
        by_view = groupby(subset, lambda op: self.opinfo.get(op).v_base)
        if self.only_merge_ops_with_view:
            if None in by_view:
                # If an op has no views, v_base will be None.
                # If we're only merging views, then we get rid of this subset.
                del by_view[None]

            for view_subset in by_view.values():
                if len(view_subset) > 1:
                    self.perform_merges_for_view_subset(view_subset)
        elif None in by_view and len(by_view[None]) > 1:
            self.perform_merges_for_view_subset(by_view[None])

    def perform_merges_for_view_subset(self, subset):
        """Perform merges for a subset of operators with the same view base.

        Parameters
        ----------
        subset : list
            Subset of operators. These need to have the same view base (can be
            None if it is None for all) for their first signal in
            ``all_signals``.
        """

        # Sort to have sequential memory.
        offsets = np.array(
            [self.opinfo.get(op).v_offset for op in subset], dtype=rc.float_dtype
        )
        sort_indices = np.argsort(offsets)
        offsets = offsets[sort_indices]
        sorted_subset = [subset[i] for i in sort_indices]

        for i, op in enumerate(sorted_subset):
            if op in self.merged:
                # Cannot merge merged operator again until dependency graph
                # has been updated
                continue

            if op in self.merged_dependents or any(
                o in self.merged for o in self.dependents[op]
            ):
                continue

            tomerge = OpsToMerge(
                op, self.merged, self.merged_dependents, self.dependents
            )

            # For a merge to be possible the view of the next operator has to
            # start where the view of op ends. Because we have sorted the
            # operators by the start of their views we can do a binary search
            # and potentially skip a number of operators at the beginning.
            start = np.searchsorted(
                offsets, offsets[i] + self.opinfo.get(op).v_size, side="left"
            )

            for op2 in sorted_subset[start:]:

                if tomerge.not_sequential(op2):
                    # If this check is true the view of op2 does not
                    # immediately follow the view of the operators being
                    # merged. Because we iterate over the operators sorted by
                    # view offset there will be a gap between the end of the
                    # operators being merged and all remaining operators to
                    # loop over. With such a gap a merge is never possible and
                    # we can cut the loop short.
                    break

                if op2 in self.merged:
                    continue

                if OpMerger.is_mergeable(op2, tomerge):
                    tomerge.add(op2)

            if len(tomerge.ops) > 1:
                self.merge(tomerge)
            elif self.only_merge_ops_with_view:
                self.might_merge.remove(op)

    def merge(self, tomerge):
        """Merges the given operators.

        This method will also update ``op_replacements``, ``sig_replacements``,
        and the internal list of merged operators to prevent further merges
        on the same operators before all required operators and signals have
        been replaced.
        """
        merged_op, merged_sig = OpMerger.merge(tomerge.ops)
        self.dg.merge(tomerge.ops, merged_op)

        # Update tracking what has been merged and might be mergeable later
        self.might_merge.difference_update(tomerge.ops)
        self.might_merge.add(merged_op)
        self.merged.update(tomerge.ops)
        self.merged_dependents.update(tomerge.all_dependents)

        for op in tomerge.ops:
            # Mark all operators referencing the same signals as merged
            # (even though they are not) to prevent them from getting
            # merged before their signals have been updated.
            for s in op.all_signals:
                self.merged.update(self.sig2ops[s])

        # Signal related updates
        self.resolve_views_on_replaced_signals(merged_sig)
        self.sig_replacements.update(merged_sig)
        self.replace_op_signals(merged_sig)
        self.update_signal_indexing(merged_op, merged_sig)

    def resolve_views_on_replaced_signals(self, replaced_signals):
        for sig in list(replaced_signals):
            for view in self.base2views[sig]:
                if view is sig:
                    continue
                assert view.base is sig
                base_replacement = replaced_signals[sig]
                offset = view.offset
                strides = tuple(
                    a // b * c
                    for a, b, c in zip_longest(
                        view.strides,
                        view.base.strides,
                        base_replacement.strides,
                        fillvalue=1,
                    )
                )
                if base_replacement.is_view:
                    offset += base_replacement.offset
                    base_replacement = base_replacement.base
                buf = base_replacement.initial_value
                initial_value = np.ndarray(
                    buffer=buf,
                    dtype=view.dtype,
                    shape=view.shape,
                    offset=offset,
                    strides=strides,
                )
                replaced_signals[view] = Signal(
                    initial_value,
                    name=view.name,
                    base=base_replacement,
                    readonly=view.readonly,
                    offset=offset,
                )

    def replace_op_signals(self, replaced_signals):
        ops = (op for s in replaced_signals for op in self.sig2ops[s])
        for v in ops:
            # Update the op's signals
            v.sets = [replaced_signals.get(s, s) for s in v.sets]
            v.incs = [replaced_signals.get(s, s) for s in v.incs]
            v.reads = [replaced_signals.get(s, s) for s in v.reads]
            v.updates = [replaced_signals.get(s, s) for s in v.updates]

    def update_signal_indexing(self, merged_op, replaced_signals):
        for s in merged_op.all_signals:
            self.sig2ops[s].add(merged_op)
            if s.is_view:
                self.base2views[s.base].add(s)

        for from_sig, to_sig in replaced_signals.items():
            self.sig2ops[to_sig] = self.sig2ops[from_sig]
            if to_sig.is_view:
                self.base2views[to_sig.base].add(to_sig)


class OpInfo:
    """Analyze and store extra information about operators."""

    _OpDetails = namedtuple(
        "_OpDetails", ["first_view", "v_offset", "v_size", "v_base"]
    )

    def __init__(self):
        super().__init__()
        self.info = {}

    def get(self, op):
        if op not in self.info:
            try:
                first_view = next(s for s in op.all_signals if s.is_view)
                self.info[op] = self._OpDetails(
                    first_view=first_view,
                    v_offset=first_view.offset,
                    v_size=first_view.nbytes,
                    v_base=first_view.base,
                )
            except StopIteration:
                self.info[op] = self._OpDetails(
                    first_view=None, v_offset=0, v_size=0, v_base=None
                )
        return self.info[op]

    def clear(self):
        self.info.clear()


class OpsToMerge:
    """Analyze and store extra information about a list of ops to be merged."""

    def __init__(self, initial_op, merged, merged_dependents, dependents):
        self.merged = merged
        self.merged_dependents = merged_dependents
        self.dependents = dependents
        self.ops = [initial_op]
        self.optype = type(initial_op)
        self.opinfo = OpInfo()

        self.all_signals = set(initial_op.all_signals)
        self.all_dependents = set(self.dependents[initial_op])

    @property
    def last_op(self):
        return self.ops[-1]

    def add(self, op):
        self.ops.append(op)
        self.all_signals.update(op.all_signals)
        self.all_dependents.update(self.dependents[op])

    @staticmethod
    def check_signals(op1, op2):
        for s1, s2 in zip(op1.all_signals, op2.all_signals):
            # If one op's signal is a view, the other must be as well
            if s1.is_view is not s2.is_view:
                return False

            if s1.is_view:
                # Views must be on the same base
                if s1.base is not s2.base:
                    return False
                # Views must have the same strides
                elif s1.strides != s2.strides:
                    return False

        return True

    def not_sequential(self, op):
        lastop = self.opinfo.get(self.ops[-1])
        return lastop.v_offset + lastop.v_size < self.opinfo.get(op).v_offset


class OpMerger:
    """Manages the op merge classes known to the optimizer."""

    mergers = {}

    @classmethod
    def is_mergeable(cls, op, tomerge):
        merger = cls.mergers[tomerge.optype]

        independent_of_ops_tomerge = (
            op not in tomerge.all_dependents
            and tomerge.dependents[op].isdisjoint(tomerge.ops)
        )
        independent_of_prior_merges = (
            op not in tomerge.merged
            and op not in tomerge.merged_dependents
            and all(o not in tomerge.merged for o in tomerge.dependents[op])
        )
        return (
            type(op) is tomerge.optype
            and independent_of_ops_tomerge
            and independent_of_prior_merges
            and cls.is_type_mergeable(tomerge.optype)
            and tomerge.check_signals(tomerge.ops[-1], op)
            and merger.check_signals(op, tomerge)
            and merger.is_mergeable(tomerge.last_op, op)
        )

    @classmethod
    def is_type_mergeable(cls, optype):
        return optype in cls.mergers

    @classmethod
    def merge(cls, ops):
        return cls.mergers[type(ops[0])].merge(ops)

    @classmethod
    def register(cls, optype):
        def register(merger):
            if optype in cls.mergers:
                warnings.warn(f"Merger for operator type {optype} overwritten.")
            cls.mergers[optype] = merger
            return merger

        return register


class Merger:
    """Base class for all op merge classes."""

    @staticmethod
    def check_signals(op, tomerge):
        return tomerge.all_signals.isdisjoint(op.all_signals)

    @staticmethod
    def is_mergeable(op1, op2):
        raise NotImplementedError("Subclasses must implement `is_mergeable`")

    @staticmethod
    def merge(ops):
        raise NotImplementedError("Cannot merge arbitrary ops.")

    @staticmethod
    def merge_dicts(*dicts):
        """Merges the given dictionaries into a single dictionary.

        This function assumes and enforces that no keys overlap.
        """
        d = {}
        for other_d in dicts:
            assert all(k not in d for k in other_d)
            d.update(other_d)
        return d


@OpMerger.register(Reset)
class ResetMerger(Merger):
    """Merge `.Reset` ops."""

    @staticmethod
    def is_mergeable(op1, op2):
        return SigMerger.check([op1.dst, op2.dst]) and op1.value == op2.value

    @staticmethod
    def merge(ops):
        dst, replacements = SigMerger.merge([o.dst for o in ops])
        return Reset(dst, ops[0].value), replacements


@OpMerger.register(Copy)
class CopyMerger(Merger):
    """Merge `.Copy` ops."""

    @staticmethod
    def is_mergeable(op1, op2):
        return (
            SigMerger.check([op1.src, op2.src])
            and SigMerger.check([op1.dst, op2.dst])
            and op1.src_slice is None
            and op1.dst_slice is None
            and op2.src_slice is None
            and op2.dst_slice is None
            and op1.inc == op2.inc
        )

    @staticmethod
    def merge_slice(signals, slices):
        if all(s is None for s in slices):
            return None
        elif any(s is None for s in slices):
            raise ValueError("Mixed Ellipsis with list of indices.")

        offset = 0
        merged_slice = []
        for sig, sl in zip(signals, slices):
            assert isinstance(sl, list), f"Expecting a list of indices, got {sl}"
            merged_slice.extend([i + offset for i in sl])
            offset += sig.size
        return merged_slice

    @staticmethod
    def merge(ops):
        src_sigs = [o.src for o in ops]
        dst_sigs = [o.dst for o in ops]

        src, src_sigr = SigMerger.merge(src_sigs)
        dst, dst_sigr = SigMerger.merge(dst_sigs)
        src_slice = CopyMerger.merge_slice(src_sigs, [o.src_slice for o in ops])
        dst_slice = CopyMerger.merge_slice(dst_sigs, [o.dst_slice for o in ops])
        return (
            Copy(src, dst, src_slice=src_slice, dst_slice=dst_slice, inc=ops[0].inc),
            Merger.merge_dicts(src_sigr, dst_sigr),
        )


@OpMerger.register(ElementwiseInc)
class ElementwiseIncMerger(Merger):
    """Merge `.ElementwiseInc` ops."""

    @staticmethod
    def is_mergeable(op1, op2):
        scalar_mult = op1.A.shape == (1,) and op2.A.shape == (1,)
        non_scalar_mult = op1.A.shape != (1,) and op2.A.shape != (1,)
        return (
            SigMerger.check([op1.X, op2.X], axis=op1.X.ndim - 1)
            and SigMerger.check([op1.Y, op2.Y], axis=op1.Y.ndim - 1)
            and (
                (scalar_mult and op1.A.initial_value == op2.A.initial_value)
                or (
                    non_scalar_mult
                    and SigMerger.check([op1.A, op2.A], axis=op1.A.ndim - 1)
                )
            )
        )

    @staticmethod
    def merge(ops):
        if all(o.A.shape == (1,) for o in ops):
            assert all(o.A.initial_value == ops[0].A.initial_value for o in ops)
            A, A_sigr = ops[0].A, {}
        else:
            A, A_sigr = SigMerger.merge([o.A for o in ops], axis=ops[0].A.ndim - 1)
        X, X_sigr = SigMerger.merge([o.X for o in ops], axis=ops[0].X.ndim - 1)
        Y, Y_sigr = SigMerger.merge([o.Y for o in ops], axis=ops[0].Y.ndim - 1)
        return (
            ElementwiseInc(A, X, Y),
            Merger.merge_dicts(A_sigr, X_sigr, Y_sigr),
        )


@OpMerger.register(DotInc)
class DotIncMerger(Merger):
    """Merge `.DotInc` ops."""

    @staticmethod
    def check_signals(op, tomerge):
        none_shared = Merger.check_signals(op, tomerge) and (
            len(tomerge.ops) < 2 or tomerge.ops[0].X is not tomerge.ops[1].X
        )
        all_x_shared = (
            op.A not in tomerge.all_signals
            and op.Y not in tomerge.all_signals
            and all(op.X not in [o.A, o.Y] and op.X is o.X for o in tomerge.ops)
        )
        return none_shared or all_x_shared

    @staticmethod
    def is_mergeable(op1, op2):
        if op1.X is op2.X:
            # simple merge might be possible
            return SigMerger.check([op1.Y, op2.Y]) and SigMerger.check([op1.A, op2.A])

        # check if BSR merge is possible
        try:
            # Not using check() for A, because A must not be a view.
            SigMerger.check_signals([op1.A, op2.A])
            from scipy.sparse import (  # pylint: disable=import-outside-toplevel
                bsr_matrix,
            )

            assert bsr_matrix
        except ImportError:
            warnings.warn(
                "Skipping some optimization steps because SciPy is "
                "not installed. Installing SciPy may result in "
                "faster simulations."
            )
            return False
        except ValueError:
            return False
        return (
            SigMerger.check([op1.X, op2.X])
            and SigMerger.check([op1.Y, op2.Y])
            and op1.A.shape != (1,)
            and op1.A.shape == op2.A.shape
        )

    @staticmethod
    def merge(ops):
        # Simple merge if all X are the same.
        if all(o.X is ops[0].X for o in ops):
            A, A_sigr = SigMerger.merge([o.A for o in ops])
            Y, Y_sigr = SigMerger.merge([o.Y for o in ops])
            return DotInc(A, ops[0].X, Y), Merger.merge_dicts(A_sigr, Y_sigr)

        assert all(o1.X is not o2.X for i, o1 in enumerate(ops) for o2 in ops[i + 1 :])

        # BSR merge if X differ
        X, X_sigr = SigMerger.merge([o.X for o in ops])
        Y, Y_sigr = SigMerger.merge([o.Y for o in ops])

        # Construct sparse A representation
        data = np.array([o.A.initial_value for o in ops], dtype=rc.float_dtype)
        if data.ndim == 1:
            raise NotImplementedError("A.ndim should be > 2")
        elif data.ndim == 2:
            raise NotImplementedError("A.ndim should be > 2")
        indptr = np.arange(len(ops) + 1, dtype=rc.int_dtype)
        indices = np.arange(len(ops), dtype=rc.int_dtype)
        name = f"bsr_merged<{ops[0].A.name}, ..., {ops[-1].A.name}>"
        readonly = all(o.A.readonly for o in ops)
        A = Signal(data, name=name, readonly=readonly)
        A_sigr = {}
        for i, s in enumerate(o.A for o in ops):
            A_sigr[s] = Signal(
                data[i],
                name=f"{s.name}[{i}]",
                base=A,
                offset=i * A.itemsize * np.prod(A.shape[1:]),
                readonly=readonly,
            )
            assert np.allclose(
                s.initial_value, A_sigr[s].initial_value, atol=0, rtol=0, equal_nan=True
            )
            assert s.shape == A_sigr[s].shape or (
                s.shape == () and A_sigr[s].shape == (1, 1)
            )

        reshape = reshape_dot(
            ops[0].A.initial_value,
            ops[0].X.initial_value,
            ops[0].Y.initial_value,
            tag=ops[0].tag,
        )
        return (
            BsrDotInc(A, X, Y, indices=indices, indptr=indptr, reshape=reshape),
            Merger.merge_dicts(X_sigr, Y_sigr, A_sigr),
        )


@OpMerger.register(SimNeurons)
class SimNeuronsMerger(Merger):
    """Merge `.SimNeurons` ops."""

    @staticmethod
    def is_mergeable(op1, op2):
        return op1.neurons == op2.neurons and all(
            SigMerger.check(s) for s in zip(op1.all_signals, op2.all_signals)
        )

    @staticmethod
    def merge(ops):
        J, J_sigr = SigMerger.merge([op.J for op in ops])
        output, out_sigr = SigMerger.merge([op.output for op in ops])
        state = {}
        state_sigr = {}
        for key in ops[0].state_idxs:
            st, st_sigr = SigMerger.merge([op.sets[op.state_idxs[key]] for op in ops])
            state[key] = st
            state_sigr.update(st_sigr)
        state.update(ops[0].state_extra)
        if any(len(op.state_extra) > 0 for op in ops[1:]):
            warnings.warn(
                "Extra state has been modified when merging two or more SimNeurons "
                f"ops associated with '{type(ops[0].neurons).__name__}' neuron types. "
                "If this causes issues, turn off the optimizer."
            )
        return (
            SimNeurons(ops[0].neurons, J, output, state=state),
            Merger.merge_dicts(J_sigr, out_sigr, state_sigr),
        )


class SigMerger:
    """Merge signals."""

    @staticmethod
    def check(signals, axis=0):
        """Checks that all signals can be concatenated along a given axis.

        For views, this includes also a check that the signals have a common
        base and agree on the strides.

        In comparison to the ``check_*`` functions, this function  does
        not throw exceptions and allows for either signals or signal views.
        """
        if len(set(signals)) != len(signals):
            # Signal appears twice in list.
            return False

        if all(s.is_view for s in signals):
            try:
                SigMerger.check_views(signals, axis=axis)
            except ValueError:
                return False
        elif all(not s.is_view for s in signals):
            try:
                SigMerger.check_signals(signals, axis=axis)
            except ValueError:
                return False
        else:
            # Mix of signals and not signals
            return False

        # If we haven't failed yet, then the signals are compatible
        return True

    @staticmethod
    def check_signals(signals, axis=0):
        """Checks that all signals can be merged along a given axis.

        If this is not possible, or any signals are views, a
        ``ValueError`` will be raised.
        """
        if any(s.is_view for s in signals):
            raise ValueError("Cannot merge views.")

        if any(s.sparse for s in signals):
            raise ValueError("Cannot merge sparse Signals")

        for s in signals:
            if s.ndim != signals[0].ndim:
                raise ValueError("Signals must have the same number of dimensions.")
            if s.ndim <= 0 and s.initial_value != signals[0].initial_value:
                raise ValueError("0-d signals must have the same initial value.")
            if (
                s.shape[:axis] != signals[0].shape[:axis]
                or s.shape[axis + 1 :] != signals[0].shape[axis + 1 :]
            ):
                raise ValueError(
                    "Signals must have same shape except on concatenation axis."
                )
            if s.dtype is not signals[0].dtype:
                raise ValueError("Signals must have the same dtype.")

    @staticmethod
    def check_views(signals, axis=0):
        """Checks that all signal views can be merged along a given axis.

        If this is not possible, or any signals are not views,
        a ``ValueError`` will be raised.

        ``signals`` must be ordered by the offset into the base signal.
        """
        if any(not s.is_view for s in signals):
            raise ValueError("Cannot merge non-views.")

        start = signals[0].offset
        for s in signals:
            if s.base is not signals[0].base:
                raise ValueError("Signals must share the same base.")
            if s.ndim != signals[0].ndim:
                raise ValueError("Signals must have the same number of dimensions.")
            if s.strides != signals[0].strides:
                raise ValueError("Signals must have equal strides.")
            if (
                s.shape[:axis] != signals[0].shape[:axis]
                or s.shape[axis + 1 :] != signals[0].shape[axis + 1 :]
            ):
                raise ValueError(
                    "Signals must have same shape except on concatenation axis."
                )
            if s.offset != start:
                raise ValueError("Views are not sequential.")
            start = s.offset + s.nbytes

    @staticmethod
    def merge(signals, axis=0):
        """Merges multiple signals or signal views into one contiguous signal.

        Note that if any of the signals are linked to another signal (by being
        the base of a view), the merged signal will not reflect those links
        anymore.

        Parameters
        ----------
        signals : sequence
            Signals to merge. Must not contain views.
        axis : int, optional
            Axis along which to concatenate the signals.

        Returns
        -------
        merged_signal : Signal
            The merged signal.
        replacements : dict
            Dictionary mapping from the old signals to new  signals that are
            a view into the merged signal. Used to replace old signals.
        """
        are_views = [s.is_view for s in signals]
        if all(are_views):
            return SigMerger.merge_views(signals, axis=axis)
        elif not any(are_views):
            return SigMerger.merge_signals(signals, axis=axis)
        else:
            raise ValueError("Cannot merge mixed views and non-views.")

    @staticmethod
    def merge_signals(signals, axis=0):
        """Merges multiple signal into one contiguous signal.

        Note that if any of the signals are linked to another signal (by being
        the base of a view), the merged signal will not reflect
        those links anymore.

        Parameters
        ----------
        signals : sequence
            Signals to merge. Must not contain views.
        axis : int, optional
            Axis along which to concatenate the signals.

        Returns
        -------
        merged_signal : Signal
            The merged signal.
        replacements : dict
            Dictionary mapping from the old signals to new  signals that are
            a view into the merged signal. Used to replace old signals.
        """
        SigMerger.check_signals(signals, axis=axis)

        if signals[0].ndim > 0:
            initial_value = np.concatenate(
                [s.initial_value for s in signals], axis=axis
            )
        else:
            initial_value = signals[0].initial_value
        readonly = all(s.readonly for s in signals)
        name = "merged<" + signals[0].name + ", ..., " + signals[-1].name + ">"
        merged_signal = Signal(initial_value, name=name, readonly=readonly)

        if signals[0].ndim > 0:
            replacements = {}
            start = 0
            for s in signals:
                size = s.shape[axis]
                indexing = [slice(None)] * initial_value.ndim
                indexing[axis] = slice(start, start + size)
                replacements[s] = merged_signal[tuple(indexing)]
                start += size
        else:
            replacements = {s: merged_signal for s in signals}

        return merged_signal, replacements

    @staticmethod
    def merge_views(signals, axis=0):
        """Merges multiple signal views into one contiguous signal view.

        Parameters
        ----------
        signals : sequence
            Signals to merge. Must only contain views.
        axis : int, optional
            Axis along which to concatenate the signals.

        Returns
        -------
        merged_signal : Signal
            The merged signal.
        replacements : dict
            Dictionary mapping from the old signals to new  signals that are
            a view into the merged signal. Used to replace old signals.
        """
        SigMerger.check_views(signals, axis=axis)

        # abs_offset = min(s.abs_offset for s in signals)
        offset = min(s.offset for s in signals)
        shape = (
            signals[0].shape[:axis]
            + (sum(s.shape[axis] for s in signals),)
            + signals[0].shape[axis + 1 :]
        )
        initial_value = np.ndarray(
            buffer=signals[0].base.initial_value,
            dtype=signals[0].dtype,
            shape=shape,
            offset=offset,
            strides=signals[0].strides,
        )
        merged_signal = Signal(
            initial_value,
            name=signals[0].base.name,
            base=signals[0].base,
            readonly=all(s.readonly for s in signals),
            offset=offset,
        )

        return merged_signal, {}


def groupby(lst, keyfunc=lambda item: item):
    """Groups the given list by the value returned by ``keyfunc``.

    Similar to ``itertools.groupby``, but returns a dict, and does not depend
    on the order of the input list.
    """
    d = defaultdict(list)
    for item in lst:
        d[keyfunc(item)].append(item)
    return d



import numpy as np

from nengo.connection import Connection
from nengo.exceptions import ObsoleteError
from nengo.network import Network
from nengo.networks.ensemblearray import EnsembleArray
from nengo.node import Node


class InputGatedMemory(Network):
    """Stores a given vector in memory, with input controlled by a gate.

    Parameters
    ----------
    n_neurons : int
        Number of neurons per dimension in the vector.
    dimensions : int
        Dimensionality of the vector.

    feedback : float, optional
        Strength of the recurrent connection from the memory to itself.
    difference_gain : float, optional
        Strength of the connection from the difference ensembles to the
        memory ensembles.
    recurrent_synapse : float, optional

    difference_synapse : Synapse
        If None, ...
    **kwargs
        Keyword arguments passed through to ``nengo.Network``
        like 'label' and 'seed'.

    Attributes
    ----------
    diff : EnsembleArray
        Represents the difference between the desired vector and
        the current vector represented by ``mem``.
    gate : Node
        With input of 0, the network is not gated, and ``mem`` will be updated
        to minimize ``diff``. With input greater than 0, the network will be
        increasingly gated such that ``mem`` will retain its current value,
        and ``diff`` will be inhibited.
    input : Node
        The desired vector.
    mem : EnsembleArray
        Integrative population that stores the vector.
    output : Node
        The vector currently represented by ``mem``.
    reset : Node
        With positive input, the ``mem`` population will be inhibited,
        effectively wiping out the vector currently being remembered.
    """

    def __init__(
        self,
        n_neurons,
        dimensions,
        feedback=1.0,
        difference_gain=1.0,
        recurrent_synapse=0.1,
        difference_synapse=None,
        **kwargs,
    ):

        if "net" in kwargs:
            raise ObsoleteError("The 'net' argument is no longer supported.")
        kwargs.setdefault("label", "Input gated memory")
        super().__init__(**kwargs)

        if difference_synapse is None:
            difference_synapse = recurrent_synapse

        n_total_neurons = n_neurons * dimensions

        with self:
            # integrator to store value
            self.mem = EnsembleArray(n_neurons, dimensions, label="mem")
            Connection(
                self.mem.output,
                self.mem.input,
                transform=feedback,
                synapse=recurrent_synapse,
            )

            # calculate difference between stored value and input
            self.diff = EnsembleArray(n_neurons, dimensions, label="diff")
            Connection(self.mem.output, self.diff.input, transform=-1)

            # feed difference into integrator
            Connection(
                self.diff.output,
                self.mem.input,
                transform=difference_gain,
                synapse=difference_synapse,
            )

            # gate difference (if gate==0, update stored value,
            # otherwise retain stored value)
            self.gate = Node(size_in=1)
            self.diff.add_neuron_input()
            Connection(
                self.gate,
                self.diff.neuron_input,
                transform=np.ones((n_total_neurons, 1)) * -10,
                synapse=None,
            )

            # reset input (if reset=1, remove all values, and set to 0)
            self.reset = Node(size_in=1)
            Connection(
                self.reset,
                self.mem.add_neuron_input(),
                transform=np.ones((n_total_neurons, 1)) * -3,
                synapse=None,
            )

        self.input = self.diff.input
        self.output = self.mem.output



"""
These are borrowed from SciPy and used under their license:

Copyright (c) 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright (c) 2003-2012 SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of Enthought nor the names of the SciPy Developers
     may be used to endorse or promote products derived from this software
     without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
"""
# From scipy/sparse/linalg/_onenormest.py v0.18.0

from __future__ import absolute_import, division

import numpy as np

from nengo._vendor.scipy.sparse.linalg_interface import aslinearoperator


def onenormest(A, t=2, itmax=5, compute_v=False, compute_w=False):
    """
    Compute a lower bound of the 1-norm of a sparse matrix.

    Parameters
    ----------
    A : ndarray or other linear operator
        A linear operator that can be transposed and that can
        produce matrix products.
    t : int, optional
        A positive parameter controlling the tradeoff between
        accuracy versus time and memory usage.
        Larger values take longer and use more memory
        but give more accurate output.
    itmax : int, optional
        Use at most this many iterations.
    compute_v : bool, optional
        Request a norm-maximizing linear operator input vector if True.
    compute_w : bool, optional
        Request a norm-maximizing linear operator output vector if True.

    Returns
    -------
    est : float
        An underestimate of the 1-norm of the sparse matrix.
    v : ndarray, optional
        The vector such that ||Av||_1 == est*||v||_1.
        It can be thought of as an input to the linear operator
        that gives an output with particularly large norm.
    w : ndarray, optional
        The vector Av which has relatively large 1-norm.
        It can be thought of as an output of the linear operator
        that is relatively large in norm compared to the input.

    Notes
    -----
    This is algorithm 2.4 of [1].

    In [2] it is described as follows.
    "This algorithm typically requires the evaluation of
    about 4t matrix-vector products and almost invariably
    produces a norm estimate (which is, in fact, a lower
    bound on the norm) correct to within a factor 3."

    .. versionadded:: 0.13.0

    References
    ----------
    .. [1] Nicholas J. Higham and Francoise Tisseur (2000),
           "A Block Algorithm for Matrix 1-Norm Estimation,
           with an Application to 1-Norm Pseudospectra."
           SIAM J. Matrix Anal. Appl. Vol. 21, No. 4, pp. 1185-1201.

    .. [2] Awad H. Al-Mohy and Nicholas J. Higham (2009),
           "A new scaling and squaring algorithm for the matrix exponential."
           SIAM J. Matrix Anal. Appl. Vol. 31, No. 3, pp. 970-989.

    """

    # Check the input.
    A = aslinearoperator(A)
    if A.shape[0] != A.shape[1]:
        raise ValueError('expected the operator to act like a square matrix')

    # If the operator size is small compared to t,
    # then it is easier to compute the exact norm.
    # Otherwise estimate the norm.
    n = A.shape[1]
    if t >= n:
        A_explicit = np.asarray(aslinearoperator(A).matmat(np.identity(n)))
        if A_explicit.shape != (n, n):
            raise Exception('internal error: ',
                            'unexpected shape ' + str(A_explicit.shape))
        col_abs_sums = abs(A_explicit).sum(axis=0)
        if col_abs_sums.shape != (n, ):
            raise Exception('internal error: ',
                            'unexpected shape ' + str(col_abs_sums.shape))
        argmax_j = np.argmax(col_abs_sums)
        v = elementary_vector(n, argmax_j)
        w = A_explicit[:, argmax_j]
        est = col_abs_sums[argmax_j]
    else:
        est, v, w, nmults, nresamples = _onenormest_core(A, A.H, t, itmax)

    # Report the norm estimate along with some certificates of the estimate.
    if compute_v or compute_w:
        result = (est,)
        if compute_v:
            result += (v,)
        if compute_w:
            result += (w,)
        return result
    else:
        return est


def _blocked_elementwise(func):
    """
    Decorator for an elementwise function, to apply it blockwise along
    first dimension, to avoid excessive memory usage in temporaries.
    """
    block_size = 2**20

    def wrapper(x):
        if x.shape[0] < block_size:
            return func(x)
        else:
            y0 = func(x[:block_size])
            y = np.zeros((x.shape[0],) + y0.shape[1:], dtype=y0.dtype)
            y[:block_size] = y0
            del y0
            for j in range(block_size, x.shape[0], block_size):
                y[j:j+block_size] = func(x[j:j+block_size])
            return y
    return wrapper


@_blocked_elementwise
def sign_round_up(X):
    """
    This should do the right thing for both real and complex matrices.

    From Higham and Tisseur:
    "Everything in this section remains valid for complex matrices
    provided that sign(A) is redefined as the matrix (aij / |aij|)
    (and sign(0) = 1) transposes are replaced by conjugate transposes."

    """
    Y = X.copy()
    Y[Y == 0] = 1
    Y /= np.abs(Y)
    return Y


@_blocked_elementwise
def _max_abs_axis1(X):
    return np.max(np.abs(X), axis=1)


def _sum_abs_axis0(X):
    block_size = 2**20
    r = None
    for j in range(0, X.shape[0], block_size):
        y = np.sum(np.abs(X[j:j+block_size]), axis=0)
        if r is None:
            r = y
        else:
            r += y
    return r


def elementary_vector(n, i):
    v = np.zeros(n, dtype=float)
    v[i] = 1
    return v


def vectors_are_parallel(v, w):
    # Columns are considered parallel when they are equal or negative.
    # Entries are required to be in {-1, 1},
    # which guarantees that the magnitudes of the vectors are identical.
    if v.ndim != 1 or v.shape != w.shape:
        raise ValueError('expected conformant vectors with entries in {-1,1}')
    n = v.shape[0]
    return np.dot(v, w) == n


def every_col_of_X_is_parallel_to_a_col_of_Y(X, Y):
    for v in X.T:
        if not any(vectors_are_parallel(v, w) for w in Y.T):
            return False
    return True


def column_needs_resampling(i, X, Y=None):
    # column i of X needs resampling if either
    # it is parallel to a previous column of X or
    # it is parallel to a column of Y
    n, t = X.shape
    v = X[:, i]
    if any(vectors_are_parallel(v, X[:, j]) for j in range(i)):
        return True
    if Y is not None:
        if any(vectors_are_parallel(v, w) for w in Y.T):
            return True
    return False


def resample_column(i, X):
    X[:, i] = np.random.randint(0, 2, size=X.shape[0])*2 - 1


def less_than_or_close(a, b):
    return np.allclose(a, b) or (a < b)


def _algorithm_2_2(A, AT, t):  # noqa: C901
    """
    This is Algorithm 2.2.

    Parameters
    ----------
    A : ndarray or other linear operator
        A linear operator that can produce matrix products.
    AT : ndarray or other linear operator
        The transpose of A.
    t : int, optional
        A positive parameter controlling the tradeoff between
        accuracy versus time and memory usage.

    Returns
    -------
    g : sequence
        A non-negative decreasing vector
        such that g[j] is a lower bound for the 1-norm
        of the column of A of jth largest 1-norm.
        The first entry of this vector is therefore a lower bound
        on the 1-norm of the linear operator A.
        This sequence has length t.
    ind : sequence
        The ith entry of ind is the index of the column A whose 1-norm
        is given by g[i].
        This sequence of indices has length t, and its entries are
        chosen from range(n), possibly with repetition,
        where n is the order of the operator A.

    Notes
    -----
    This algorithm is mainly for testing.
    It uses the 'ind' array in a way that is similar to
    its usage in algorithm 2.4.  This algorithm 2.2 may be easier to test,
    so it gives a chance of uncovering bugs related to indexing
    which could have propagated less noticeably to algorithm 2.4.

    """
    A_linear_operator = aslinearoperator(A)
    AT_linear_operator = aslinearoperator(AT)
    n = A_linear_operator.shape[0]

    # Initialize the X block with columns of unit 1-norm.
    X = np.ones((n, t))
    if t > 1:
        X[:, 1:] = np.random.randint(0, 2, size=(n, t-1))*2 - 1
    X /= float(n)

    # Iteratively improve the lower bounds.
    # Track extra things, to assert invariants for debugging.
    g_prev = None
    h_prev = None
    k = 1
    ind = range(t)
    while True:
        Y = np.asarray(A_linear_operator.matmat(X))
        g = _sum_abs_axis0(Y)
        best_j = np.argmax(g)
        g.sort()
        g = g[::-1]
        S = sign_round_up(Y)
        Z = np.asarray(AT_linear_operator.matmat(S))
        h = _max_abs_axis1(Z)

        # If this algorithm runs for fewer than two iterations,
        # then its return values do not have the properties indicated
        # in the description of the algorithm.
        # In particular, the entries of g are not 1-norms of any
        # column of A until the second iteration.
        # Therefore we will require the algorithm to run for at least
        # two iterations, even though this requirement is not stated
        # in the description of the algorithm.
        if k >= 2:
            if less_than_or_close(max(h), np.dot(Z[:, best_j], X[:, best_j])):
                break
        ind = np.argsort(h)[::-1][:t]
        h = h[ind]
        for j in range(t):
            X[:, j] = elementary_vector(n, ind[j])

        # Check invariant (2.2).
        if k >= 2:
            if not less_than_or_close(g_prev[0], h_prev[0]):
                raise Exception('invariant (2.2) is violated')
            if not less_than_or_close(h_prev[0], g[0]):
                raise Exception('invariant (2.2) is violated')

        # Check invariant (2.3).
        if k >= 3:
            for j in range(t):
                if not less_than_or_close(g[j], g_prev[j]):
                    raise Exception('invariant (2.3) is violated')

        # Update for the next iteration.
        g_prev = g
        h_prev = h
        k += 1

    # Return the lower bounds and the corresponding column indices.
    return g, ind


def _onenormest_core(A, AT, t, itmax):  # noqa: C901
    """
    Compute a lower bound of the 1-norm of a sparse matrix.

    Parameters
    ----------
    A : ndarray or other linear operator
        A linear operator that can produce matrix products.
    AT : ndarray or other linear operator
        The transpose of A.
    t : int, optional
        A positive parameter controlling the tradeoff between
        accuracy versus time and memory usage.
    itmax : int, optional
        Use at most this many iterations.

    Returns
    -------
    est : float
        An underestimate of the 1-norm of the sparse matrix.
    v : ndarray, optional
        The vector such that ||Av||_1 == est*||v||_1.
        It can be thought of as an input to the linear operator
        that gives an output with particularly large norm.
    w : ndarray, optional
        The vector Av which has relatively large 1-norm.
        It can be thought of as an output of the linear operator
        that is relatively large in norm compared to the input.
    nmults : int, optional
        The number of matrix products that were computed.
    nresamples : int, optional
        The number of times a parallel column was observed,
        necessitating a re-randomization of the column.

    Notes
    -----
    This is algorithm 2.4.

    """
    # This function is a more or less direct translation
    # of Algorithm 2.4 from the Higham and Tisseur (2000) paper.
    A_linear_operator = aslinearoperator(A)
    AT_linear_operator = aslinearoperator(AT)
    if itmax < 2:
        raise ValueError('at least two iterations are required')
    if t < 1:
        raise ValueError('at least one column is required')
    n = A.shape[0]
    if t >= n:
        raise ValueError('t should be smaller than the order of A')
    # Track the number of big*small matrix multiplications
    # and the number of resamplings.
    nmults = 0
    nresamples = 0
    # "We now explain our choice of starting matrix.  We take the first
    # column of X to be the vector of 1s [...] This has the advantage that
    # for a matrix with nonnegative elements the algorithm converges
    # with an exact estimate on the second iteration, and such matrices
    # arise in applications [...]"
    X = np.ones((n, t), dtype=float)
    # "The remaining columns are chosen as rand{-1,1},
    # with a check for and correction of parallel columns,
    # exactly as for S in the body of the algorithm."
    if t > 1:
        for i in range(1, t):
            # These are technically initial samples, not resamples,
            # so the resampling count is not incremented.
            resample_column(i, X)
        for i in range(t):
            while column_needs_resampling(i, X):
                resample_column(i, X)
                nresamples += 1
    # "Choose starting matrix X with columns of unit 1-norm."
    X /= float(n)
    # "indices of used unit vectors e_j"
    ind_hist = np.zeros(0, dtype=np.intp)
    est_old = 0
    S = np.zeros((n, t), dtype=float)
    k = 1
    ind = None
    while True:
        Y = np.asarray(A_linear_operator.matmat(X))
        nmults += 1
        mags = _sum_abs_axis0(Y)
        est = np.max(mags)
        best_j = np.argmax(mags)
        if est > est_old or k == 2:
            if k >= 2:
                ind_best = ind[best_j]
            w = Y[:, best_j]
        # (1)
        if k >= 2 and est <= est_old:
            est = est_old
            break
        est_old = est
        S_old = S
        if k > itmax:
            break
        S = sign_round_up(Y)
        del Y
        # (2)
        if every_col_of_X_is_parallel_to_a_col_of_Y(S, S_old):
            break
        if t > 1:
            # "Ensure that no column of S is parallel to another column of S
            # or to a column of S_old by replacing columns of S by rand{-1,1}."
            for i in range(t):
                while column_needs_resampling(i, S, S_old):
                    resample_column(i, S)
                    nresamples += 1
        del S_old
        # (3)
        Z = np.asarray(AT_linear_operator.matmat(S))
        nmults += 1
        h = _max_abs_axis1(Z)
        del Z
        # (4)
        if k >= 2 and max(h) == h[ind_best]:
            break
        # "Sort h so that h_first >= ... >= h_last
        # and re-order ind correspondingly."
        #
        # Later on, we will need at most t+len(ind_hist) largest
        # entries, so drop the rest
        ind = np.argsort(h)[::-1][:t+len(ind_hist)].copy()
        del h
        if t > 1:
            # (5)
            # Break if the most promising t vectors have been visited already.
            if np.in1d(ind[:t], ind_hist).all():
                break
            # Put the most promising unvisited vectors at the front of the list
            # and put the visited vectors at the end of the list.
            # Preserve the order of the indices induced by the ordering of h.
            seen = np.in1d(ind, ind_hist)
            ind = np.concatenate((ind[~seen], ind[seen]))
        for j in range(t):
            X[:, j] = elementary_vector(n, ind[j])

        new_ind = ind[:t][~np.in1d(ind[:t], ind_hist)]
        ind_hist = np.concatenate((ind_hist, new_ind))
        k += 1
    v = elementary_vector(n, ind_best)
    return est, v, w, nmults, nresamples



import os

# The actual tests will execute the code anyhow so the following code can
# safely be ignored from the coverage tests
if os.name == 'nt':  # pragma: no cover
    import msvcrt

    LOCK_EX = 0x1
    LOCK_SH = 0x2
    LOCK_NB = 0x4
    LOCK_UN = msvcrt.LK_UNLCK

elif os.name == 'posix':  # pragma: no cover
    import fcntl

    LOCK_EX = fcntl.LOCK_EX
    LOCK_SH = fcntl.LOCK_SH
    LOCK_NB = fcntl.LOCK_NB
    LOCK_UN = fcntl.LOCK_UN

else:  # pragma: no cover
    raise RuntimeError('PortaLocker only defined for nt and posix platforms')




import numpy as np
import pytest

import nengo
from nengo import params
from nengo.exceptions import ConfigError, ObsoleteError, ValidationError
from nengo.params import FunctionInfo


def test_default():
    """A default value is immediately available, but can be overridden."""

    class Test:
        p = params.Parameter("p", default=1)

    inst1 = Test()
    inst2 = Test()
    assert inst1.p == 1
    assert inst2.p == 1
    inst1.p = "a"
    assert inst1.p == "a"
    assert inst2.p == 1


def test_optional():
    """Optional Parameters can bet set to None."""

    class Test:
        m = params.Parameter("m", default=1, optional=False)
        o = params.Parameter("o", default=1, optional=True)

    inst = Test()
    with pytest.raises(ValidationError, match="Parameter is not optional"):
        inst.m = None
    assert inst.m == 1
    inst.o = None
    assert inst.o is None


def test_readonly():
    """Readonly Parameters can only be set once."""

    class Test:
        p = params.Parameter("p", default=1, readonly=False)
        r = params.Parameter("r", default=None, readonly=True)

    inst = Test()
    assert inst.p == 1
    assert inst.r is None
    inst.p = 2
    inst.r = "set"
    assert inst.p == 2
    assert inst.r == "set"
    inst.p = 3
    with pytest.raises(ValidationError, match="is read-only and cannot be changed"):
        inst.r = "set again"
    assert inst.p == 3
    assert inst.r == "set"


def test_obsoleteparam():
    """ObsoleteParams must not be set."""

    class Test:
        ab = params.ObsoleteParam("ab", "msg")

    inst = Test()

    # cannot be read
    with pytest.raises(ObsoleteError):
        print(inst.ab)

    # can only be assigned Unconfigurable
    inst.ab = params.Unconfigurable
    with pytest.raises(ObsoleteError):
        inst.ab = True


def test_boolparam():
    """BoolParams can only be booleans."""

    class Test:
        bp = params.BoolParam("bp", default=False)

    inst = Test()
    assert not inst.bp
    inst.bp = True
    assert inst.bp
    with pytest.raises(ValidationError, match="[Mm]ust be of type 'bool'"):
        inst.bp = 1


def test_numberparam():
    """NumberParams can be numbers constrained to a range."""

    class Test:
        np = params.NumberParam("np", default=1.0)
        np_l = params.NumberParam("np_l", default=1.0, low=0.0)
        np_h = params.NumberParam("np_h", default=-1.0, high=0.0)
        np_lh = params.NumberParam("np_lh", default=1.0, low=-1.0, high=1.0)

    inst = Test()

    # defaults
    assert inst.np == 1.0
    assert inst.np_l == 1.0
    assert inst.np_h == -1.0
    assert inst.np_lh == 1.0

    # respect low boundaries
    inst.np = -10
    with pytest.raises(ValidationError):
        inst.np_l = -10
    with pytest.raises(ValidationError):
        inst.np_lh = -10
    assert inst.np == -10
    assert inst.np_l == 1.0
    assert inst.np_lh == 1.0
    # equal to the low boundary is ok though!
    inst.np_lh = -1.0
    assert inst.np_lh == -1.0

    # respect high boundaries
    inst.np = 10
    with pytest.raises(ValidationError):
        inst.np_h = 10
    with pytest.raises(ValidationError):
        inst.np_lh = 10
    assert inst.np == 10
    assert inst.np_h == -1.0
    assert inst.np_lh == -1.0
    # equal to the high boundary is ok though!
    inst.np_lh = 1.0
    assert inst.np_lh == 1.0

    # ensure scalar array works
    inst.np = np.array(2.0)
    assert inst.np == 2.0

    # must be a number!
    with pytest.raises(ValidationError):
        inst.np = "a"


def test_intparam():
    """IntParams are like NumberParams but must be an int."""

    class Test:
        ip = params.IntParam("ip", default=1, low=0, high=2)

    inst = Test()
    assert inst.ip == 1
    with pytest.raises(ValidationError):
        inst.ip = -1
    with pytest.raises(ValidationError):
        inst.ip = 3
    with pytest.raises(ValidationError):
        inst.ip = "a"


def test_stringparam():
    """StringParams must be strings."""

    class Test:
        sp = params.StringParam("sp", default="Hi")

    inst = Test()
    assert inst.sp == "Hi"

    inst.sp = "goodbye"
    assert inst.sp == "goodbye"

    # Bytes no good
    with pytest.raises(ValidationError):
        inst.sp = b"hello"

    # Non-strings no good
    with pytest.raises(ValidationError):
        inst.sp = 1


def test_enumparam():
    class Test:
        ep = params.EnumParam("ep", default="a", values=("a", "b", "c"))

    inst = Test()
    assert inst.ep == "a"
    inst.ep = "b"
    assert inst.ep == "b"
    inst.ep = "c"
    assert inst.ep == "c"
    inst.ep = "A"
    assert inst.ep == "a"
    with pytest.raises(ValueError):
        inst.ep = "d"
    with pytest.raises(ValueError):
        inst.ep = 3


def test_tupleparam():
    class Test:
        tp = params.TupleParam("tp", default=(0, 0, 0), optional=True)
        tp3 = params.TupleParam("tp3", default=(0, 0, 0), length=3)

    inst = Test()
    inst.tp = (1, 2, 3)
    inst.tp = (1, 2, 3, 4)
    inst.tp = None
    inst.tp3 = (1, 2, 3)
    inst.tp3 = [1.2, 2.3, 3.4]
    with pytest.raises(ValidationError):
        inst.tp = 1
    with pytest.raises(ValidationError):
        inst.tp3 = (1, 2)
    with pytest.raises(ValidationError):
        inst.tp3 = (1, 2, 3, 4)


def test_shapeparam():
    class Test:
        sp2 = params.ShapeParam("sp2", default=(0, 0), length=2, low=None)
        sp3 = params.ShapeParam("sp3", default=(0, 0, 0), length=3, optional=True)

    inst = Test()
    assert inst.sp2 == (0, 0)
    assert inst.sp3 == (0, 0, 0)
    inst.sp2 = (-1, 2)
    assert inst.sp2 == (-1, 2)
    inst.sp3 = None
    assert inst.sp3 is None
    inst.sp3 = (0, 2, 3)
    assert inst.sp3 == (0, 2, 3)
    with pytest.raises(ValidationError):
        inst.sp2 = (1, 2, 3)
    with pytest.raises(ValidationError):
        inst.sp3 = (-1, 2, 3)
    with pytest.raises(ValidationError):
        inst.sp3 = (1, 2.0, 3)


def test_dictparam():
    """DictParams must be dictionaries."""

    class Test:
        dp = params.DictParam("dp", default={"a": 1})

    inst1 = Test()
    assert inst1.dp == {"a": 1}
    inst1.dp["b"] = 2

    # The default dict is mutable -- other instances will get the same dict
    inst2 = Test()
    assert inst2.dp == {"a": 1, "b": 2}

    # Non-dicts no good
    with pytest.raises(ValidationError):
        inst2.dp = [("a", 1), ("b", 2)]


def test_ndarrayparam():
    """NdarrayParams must be able to be made into float ndarrays."""

    class Test:
        ndp = params.NdarrayParam("ndp", default=None, shape=("*",))
        ella = params.NdarrayParam("ella", default=None, shape=(3, "..."))
        ellb = params.NdarrayParam("ellb", default=None, shape=(3, "...", 2))

    inst = Test()
    inst.ndp = np.ones(10)
    assert np.all(inst.ndp == np.ones(10))
    # Dimensionality too low
    with pytest.raises(ValidationError):
        inst.ndp = 0
    # Dimensionality too high
    with pytest.raises(ValidationError):
        inst.ndp = np.ones((1, 1))
    # Must be convertible to float array
    with pytest.raises(ValidationError):
        inst.ndp = "a"

    inst.ella = np.ones((3,))
    inst.ella = np.ones((3, 1))
    inst.ella = np.ones((3, 2, 3))
    with pytest.raises(ValueError):
        inst.ella = np.ones(4)
    inst.ellb = np.ones((3, 2))
    inst.ellb = np.ones((3, 1, 2))
    with pytest.raises(ValueError):
        inst.ellb = np.ones(3)


def test_ndarrayparam_sample_shape():
    """sample_shape dictates the shape of the sample that can be set."""

    class Test:
        ndp = params.NdarrayParam("ndp", default=None, shape=[10, "d2"])
        d2 = 3
        ndp2 = params.NdarrayParam("ndp2", default=None, shape=("label",))
        label = "label"

    inst = Test()
    # Must be shape (4, 10)
    inst.ndp = np.ones((10, 3))
    assert np.all(inst.ndp == np.ones((10, 3)))
    with pytest.raises(ValidationError):
        inst.ndp = np.ones((3, 10))
    assert np.all(inst.ndp == np.ones((10, 3)))

    with pytest.raises(ValidationError, match="not yet initialized; cannot determine"):
        inst.ndp2 = (1, 2)


def test_functionparam():
    """FunctionParam must be a function, and accept one scalar argument."""

    class Test:
        fp = params.FunctionParam("fp", default=None)

    inst = Test()
    assert inst.fp is None

    inst.fp = np.sin
    assert inst.fp.function is np.sin
    assert inst.fp.size == 1

    inst.fp = FunctionInfo(np.cos, 1)
    assert inst.fp.function is np.cos
    assert inst.fp.size == 1

    # Not OK: requires two args
    with pytest.raises(ValidationError, match="function.*must accept a single.*argu"):
        inst.fp = lambda x, y: x + y

    # Not OK: not a function
    with pytest.raises(ValidationError, match="function.*must be callable"):
        inst.fp = 0


def test_iter_params_does_not_list_obsolete_params():
    class Test:
        p1 = params.IntParam("p1")
        p2 = params.IntParam("p2")
        obsolete = params.ObsoleteParam("obsolete", "not included in params")

    assert set(params.iter_params(Test())) == {"p1", "p2"}


@pytest.mark.filterwarnings("ignore:'Node.size_out' is being overwritten")
def test_configure_all_nengo_parameters():

    # make up a non-default value for the parameter
    conv_func = {
        params.BoolParam: lambda attr: not attr.default,
        params.NumberParam: lambda attr: (
            1 if attr.default is None else attr.default + 1
        ),
        params.StringParam: lambda attr: "abc",
        params.NdarrayParam: lambda attr: np.zeros([1] * len(attr.shape)),
        nengo.base.ProcessParam: lambda attr: nengo.processes.WhiteNoise(),
        nengo.node.OutputParam: lambda attr: lambda t, x=0: x + 1,
        nengo.synapses.SynapseParam: lambda attr: nengo.synapses.Alpha(0.1),
        nengo.solvers.SolverParam: lambda attr: nengo.solvers.LstsqL2nz(
            weights=isinstance(attr, nengo.connection.ConnectionSolverParam)
        ),
        nengo.connection.ConnectionFunctionParam: lambda attr: lambda x: x + 1,
        nengo.connection.ConnectionTransformParam: lambda attr: 2.0,
        nengo.learning_rules.LearningRuleTypeParam: (
            lambda attr: nengo.learning_rules.PES()
        ),
        nengo.neurons.NeuronTypeParam: lambda attr: nengo.AdaptiveLIF(),
    }

    net = nengo.Network()

    for obj in net.objects:
        for name, attr in obj.__dict__.items():
            if (
                not isinstance(attr, params.Parameter)
                or attr.default is params.Unconfigurable
            ):
                continue

            for param, func in conv_func.items():
                if isinstance(attr, param):
                    val = func(attr)
                    break
            else:
                raise NotImplementedError

            try:
                # manually set the attribute to its default
                setattr(net.config[obj], name, attr.default)

                # set it to a non-default value
                setattr(net.config[obj], name, val)
                assert getattr(net.config[obj], name) == val

            except Exception:
                print(f"Error setting {obj}.{name}")
                raise


def test_frozenobject_reprs():
    """Params should only show in repr if their values differ from defaults."""

    class TestFO(params.FrozenObject):
        a = params.NumberParam("a", default=3, readonly=True)
        b = params.NumberParam("b")

        def __init__(self, a, b=4):
            super().__init__()
            self.a = a
            self.b = b

    assert repr(TestFO(3)) == "TestFO()"
    assert repr(TestFO(2)) == "TestFO(a=2)"
    assert repr(TestFO(2, b=3)) == "TestFO(a=2, b=3)"


def test_frozenobject_missing_arg_repr():
    class TestFO(params.FrozenObject):
        a = params.NumberParam("a", default=3, readonly=True)

        def __init__(self, a, b=4):
            super().__init__()
            self.a = a

    fobj = TestFO(3)
    assert repr(fobj).startswith("<TestFO at")
    assert fobj._argreprs == "Cannot find 'b'"


def test_ndarrayparam_coerce_defaults():
    """Tests NdarrayParam coerce_defaults if shape is none"""
    params.NdarrayParam.shape = None
    assert params.NdarrayParam("name").coerce_defaults is True


def test_parameter_get_error():
    """Tests params get ValidationError"""

    class Test:
        p = params.Parameter("something", params.Unconfigurable)

    inst = Test()
    with pytest.raises(ValidationError, match="have no defaults.*ensure the value"):
        print(inst.p)


def test_parameter_set_default_error():
    """Tests params set_default ConfigError"""
    my_param = params.Parameter("something")
    with pytest.raises(ConfigError, match="Parameter.*is not configurable"):
        params.Parameter.set_default(my_param, my_param, my_param)


def test_equal_not_equatable():
    """Tests params.equal() when not equatable"""

    class Test:
        param = params.Parameter("param", default=None)

    assert not Test.param.equatable

    a = Test()
    b = Test()

    obj1, obj2 = np.array([3]), np.array([3])  # equal, but different objects
    a.param, b.param = obj1, obj1
    assert Test.param.equal(a, b)

    a.param, b.param = obj1, obj2
    assert not Test.param.equal(a, b)


def test_coerce_value_error():
    """tests to make sure ValueError is thrown with incorrect coerce usage"""

    class Test:
        o = params.Parameter("o", default=None)

    inst = Test()

    with pytest.raises(ValueError, match="is not a valid value"):
        inst.o = params.Unconfigurable


def test_parameter_arg_errors():
    with pytest.raises(ValueError, match="'name' must be a string"):
        params.Parameter(123, default=None)  # not string

    with pytest.raises(ValueError, match="'optional' must be boolean"):
        params.Parameter("o", default=None, optional="NotABool")

    with pytest.raises(ValueError, match="'readonly' must be boolean"):
        params.Parameter("o", default=None, readonly="NotABool")



from functools import partial

import numpy as np
import pytest

import nengo
import nengo.utils.numpy as npext
from nengo.connection import ConnectionSolverParam
from nengo.dists import Choice, UniformHypersphere
from nengo.exceptions import BuildError, ValidationError
from nengo.processes import Piecewise
from nengo.solvers import LstsqL2
from nengo.transforms import Dense, NoTransform
from nengo.utils.testing import signals_allclose


def test_args(AnyNeuronType, seed, rng):
    N = 10
    d1, d2 = 3, 2

    with nengo.Network(seed=seed) as model:
        model.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        A = nengo.Ensemble(N, dimensions=d1)
        B = nengo.Ensemble(N, dimensions=d2)
        nengo.Connection(
            A,
            B,
            eval_points=rng.normal(size=(500, d1)),
            synapse=0.01,
            function=np.sin,
            transform=rng.normal(size=(d2, d1)),
        )


def test_node_to_neurons(Simulator, PositiveNeuronType, plt, seed, allclose):
    N = 50

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = PositiveNeuronType()
        a = nengo.Ensemble(N, dimensions=1)
        inn = nengo.Node(output=np.sin)
        inh = nengo.Node(Piecewise({0: 0, 0.5: 1}))
        nengo.Connection(inn, a)
        nengo.Connection(inh, a.neurons, transform=[[-5]] * N)

        inn_p = nengo.Probe(inn, "output")
        a_p = nengo.Probe(a, "decoded_output", synapse=0.1)
        inh_p = nengo.Probe(inh, "output")

    with Simulator(m) as sim:
        sim.run(1.0)
    t = sim.trange()
    ideal = np.sin(t)
    ideal[t >= 0.5] = 0

    plt.plot(t, sim.data[inn_p], label="Input")
    plt.plot(t, sim.data[a_p], label="Neuron approx, synapse=0.1")
    plt.plot(t, sim.data[inh_p], label="Inhib signal")
    plt.plot(t, ideal, label="Ideal output")
    plt.legend(loc="best", fontsize="small")

    assert allclose(sim.data[a_p][-10:], 0, atol=0.1, rtol=0.01)


def test_ensemble_to_neurons(Simulator, PositiveNeuronType, plt, seed, allclose):
    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].neuron_type = PositiveNeuronType()
        ens = nengo.Ensemble(40, dimensions=1)
        inhibitor = nengo.Ensemble(40, dimensions=1)
        stim = nengo.Node(output=np.sin)
        inhibition = nengo.Node(Piecewise({0: 0, 0.5: 1}))
        nengo.Connection(stim, ens)
        nengo.Connection(inhibition, inhibitor)
        nengo.Connection(
            inhibitor, ens.neurons, transform=-10 * np.ones((ens.n_neurons, 1))
        )

        stim_p = nengo.Probe(stim, "output")
        ens_p = nengo.Probe(ens, "decoded_output", synapse=0.05)
        inhibitor_p = nengo.Probe(inhibitor, "decoded_output", synapse=0.05)
        inhibition_p = nengo.Probe(inhibition, "output")

    with Simulator(net) as sim:
        sim.run(1.0)
    t = sim.trange()
    ideal = np.sin(t)
    ideal[t >= 0.5] = 0

    plt.plot(t, sim.data[stim_p], label="Input")
    plt.plot(t, sim.data[ens_p], label="`ens` value, pstc=0.05")
    plt.plot(t, sim.data[inhibitor_p], label="`inhibitor` value, pstc=0.05")
    plt.plot(t, sim.data[inhibition_p], label="Inhibition signal")
    plt.plot(t, ideal, label="Ideal output")
    plt.legend(loc=0, prop={"size": 10})

    assert allclose(sim.data[ens_p][-10:], 0, atol=0.1, rtol=0.01)
    assert allclose(sim.data[inhibitor_p][-10:], 1, atol=0.1, rtol=0.01)


def test_node_to_ensemble(Simulator, NonDirectNeuronType, plt, seed, allclose):
    N = 50

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        input_node = nengo.Node(output=lambda t: [np.sin(t * 3), np.cos(t * 3)])
        a = nengo.Ensemble(N * 1, dimensions=1)
        b = nengo.Ensemble(N * 1, dimensions=1)
        c = nengo.Ensemble(N * 2, dimensions=2)
        d = nengo.Ensemble(N, neuron_type=nengo.Direct(), dimensions=3)

        nengo.Connection(input_node, a, function=lambda x: -x[0])
        nengo.Connection(input_node[:1], b, function=lambda x: -x)
        nengo.Connection(input_node, c, function=lambda x: -(x ** 2))
        nengo.Connection(
            input_node, d, function=lambda x: [-x[0], -(x[0] ** 2), -(x[1] ** 2)]
        )

        a_p = nengo.Probe(a, "decoded_output", synapse=0.01)
        b_p = nengo.Probe(b, "decoded_output", synapse=0.01)
        c_p = nengo.Probe(c, "decoded_output", synapse=0.01)
        d_p = nengo.Probe(d, "decoded_output", synapse=0.01)

    with Simulator(m) as sim:
        sim.run(2.0)
    t = sim.trange()

    plt.plot(t, sim.data[a_p])
    plt.plot(t, sim.data[b_p])
    plt.plot(t, sim.data[c_p])
    plt.plot(t, sim.data[d_p])
    plt.legend(
        [
            "-sin",
            "-sin",
            "-(sin ** 2)",
            "-(cos ** 2)",
            "-sin",
            "-(sin ** 2)",
            "-(cos ** 2)",
        ],
        loc="best",
        fontsize="small",
    )

    assert allclose(sim.data[a_p][-10:], sim.data[d_p][-10:][:, 0], atol=0.1, rtol=0.01)
    assert allclose(sim.data[b_p][-10:], sim.data[d_p][-10:][:, 0], atol=0.1, rtol=0.01)
    assert allclose(
        sim.data[c_p][-10:], sim.data[d_p][-10:][:, 1:3], atol=0.1, rtol=0.01
    )


def test_neurons_to_ensemble(Simulator, PositiveNeuronType, plt, seed):
    N = 20

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = PositiveNeuronType()
        a = nengo.Ensemble(N * 2, dimensions=2)
        b = nengo.Ensemble(N, dimensions=1)
        c = nengo.Ensemble(N, dimensions=N * 2)
        nengo.Connection(a.neurons, b, transform=-5 * np.ones((1, N * 2)))
        nengo.Connection(a.neurons, c)

        b_p = nengo.Probe(b, "decoded_output", synapse=0.01)
        c_p = nengo.Probe(c, "decoded_output", synapse=0.01)

    with Simulator(m) as sim:
        sim.run(0.1)
    t = sim.trange()

    plt.plot(t, sim.data[b_p], c="b")
    plt.plot(t, sim.data[c_p], c="k")
    plt.legend(
        ["Negative weights", "Neurons -> Ensemble dimensions"],
        loc="best",
        fontsize="small",
    )
    plt.xlim(right=t[-1])

    assert np.all(sim.data[b_p][-10:] < 0)


def test_neurons_to_node(Simulator, NonDirectNeuronType, plt, seed, allclose):
    N = 5

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        a = nengo.Ensemble(N, dimensions=1, encoders=np.ones((N, 1)))
        out = nengo.Node(lambda t, x: x, size_in=N)
        nengo.Connection(nengo.Node(1), a)
        nengo.Connection(a.neurons, out, synapse=None)

        a_spikes = nengo.Probe(a.neurons, synapse=0.005)
        out_p = nengo.Probe(out, synapse=0.005)

    with Simulator(m) as sim:
        sim.run(0.1)
    t = sim.trange()

    plt.subplot(2, 1, 1)
    plt.title("Activity filtered with $\\tau$ = 0.005")
    plt.ylabel("Neural activity")
    plt.plot(t, sim.data[a_spikes])
    plt.xlim(right=t[-1])
    plt.subplot(2, 1, 2)
    plt.ylabel("Node activity")
    plt.plot(t, sim.data[out_p])
    plt.xlim(right=t[-1])

    assert allclose(sim.data[a_spikes], sim.data[out_p])


def test_neurons_to_neurons(Simulator, PositiveNeuronType, plt, seed, allclose):
    N1, N2 = 50, 80

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = PositiveNeuronType()
        a = nengo.Ensemble(N1, dimensions=1)
        b = nengo.Ensemble(N2, dimensions=1)
        inp = nengo.Node(output=1)
        nengo.Connection(inp, a)
        nengo.Connection(a.neurons, b.neurons, transform=-1 * np.ones((N2, N1)))

        inp_p = nengo.Probe(inp, "output")
        a_p = nengo.Probe(a, "decoded_output", synapse=0.1)
        b_p = nengo.Probe(b, "decoded_output", synapse=0.1)

    with Simulator(m) as sim:
        sim.run(0.6)
    t = sim.trange()

    plt.plot(t, sim.data[inp_p], label="Input")
    plt.plot(t, sim.data[a_p], label="A, represents input")
    plt.plot(t, sim.data[b_p], label="B, should be 0")
    plt.ylim(top=1.1)
    plt.xlim(right=t[-1])
    plt.legend(loc="best")

    assert allclose(sim.data[a_p][-10:], 1, atol=0.1, rtol=0.01)
    assert allclose(sim.data[b_p][-10:], 0, atol=0.1, rtol=0.01)


def test_function_and_transform(Simulator, plt, seed, allclose):
    """Test using both a function and a transform"""

    model = nengo.Network(seed=seed)
    with model:
        u = nengo.Node(output=lambda t: np.sin(6 * t))
        a = nengo.Ensemble(100, 1)
        b = nengo.Ensemble(200, 2, radius=1.5)
        nengo.Connection(u, a)
        nengo.Connection(a, b, function=np.square, transform=[[1.0], [-1.0]])
        ap = nengo.Probe(a, synapse=0.03)
        bp = nengo.Probe(b, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.8)
    x = np.dot(sim.data[ap] ** 2, [[1.0, -1]]).T
    y = sim.data[bp].T

    t = sim.trange()
    plt.plot(t, x[0], "b:", label="a**2")
    plt.plot(t, x[1], "g:", label="-a**2")
    plt.plot(t, y[0], "b", label="b[0]")
    plt.plot(t, y[1], "g", label="b[1]")
    plt.legend(loc=0, prop={"size": 10})
    plt.xlim(right=t[-1])

    assert allclose(x[0], y[0], atol=0.1, rtol=0.01)
    assert allclose(x[1], y[1], atol=0.1, rtol=0.01)


def test_dist_transform(Simulator, seed, allclose):
    """Using a distribution to initialize transform."""

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Connection].transform = nengo.dists.Gaussian(0.5, 1)
        n = 300
        a = nengo.Node(output=[0] * n)
        b = nengo.Node(size_in=n + 1)
        c = nengo.Ensemble(n + 2, 10)
        d = nengo.Ensemble(n + 3, 11)

        # make a couple different types of connections to make sure that a
        # correctly sized transform is being generated
        conn1 = nengo.Connection(a, b)
        conn2 = nengo.Connection(b, c)
        conn3 = nengo.Connection(b, c.neurons)
        conn4 = nengo.Connection(b[:2], c[2])
        conn5 = nengo.Connection(c, d, solver=nengo.solvers.LstsqL2(weights=True))

    assert isinstance(conn1.transform.init, nengo.dists.Gaussian)

    with Simulator(net) as sim:
        pass

    w = sim.data[conn1].weights
    assert allclose(np.mean(w), 0.5, atol=0.01)
    assert allclose(np.std(w), 1, atol=0.01)
    assert w.shape == (n + 1, n)

    assert sim.data[conn2].weights.shape == (10, n + 1)
    assert sim.data[conn3].weights.shape == (n + 2, n + 1)
    assert sim.data[conn4].weights.shape == (1, 2)
    assert sim.data[conn5].weights.shape == (n + 3, n + 2)

    # make sure the seed works (gives us the same transform)
    with nengo.Network(seed=seed) as net:
        net.config[nengo.Connection].transform = nengo.dists.Gaussian(0.5, 1)
        a = nengo.Node(output=[0] * n)
        b = nengo.Node(size_in=n + 1)
        conn = nengo.Connection(a, b)

    with Simulator(net) as sim:
        pass
    assert allclose(w, sim.data[conn].weights)


def test_weights(Simulator, AnyNeuronType, plt, seed, allclose):
    """Tests connections using a solver with weights"""
    n1, n2 = 100, 50

    def func(t):
        return [np.sin(4 * t), np.cos(12 * t)]

    transform = np.array([[0.6, -0.4]])

    m = nengo.Network(label="test_weights", seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        u = nengo.Node(output=func)
        a = nengo.Ensemble(n1, dimensions=2, radius=1.4)
        b = nengo.Ensemble(n2, dimensions=1)
        bp = nengo.Probe(b)

        nengo.Connection(u, a)
        nengo.Connection(
            a, b, synapse=0.01, transform=transform, solver=LstsqL2(weights=True)
        )

    with Simulator(m) as sim:
        sim.run(1.0)

    t = sim.trange()
    x = np.array(func(t)).T
    y = np.dot(x, transform.T)
    z = nengo.Lowpass(0.01).filt(sim.data[bp], dt=sim.dt)
    assert signals_allclose(
        t, y, z, atol=0.15, buf=0.1, delay=0.025, plt=plt, allclose=allclose
    )


@pytest.mark.filterwarnings(
    "ignore:For connections from.*setting the solver has no effect",
    "ignore:For connections to.*setting `weights=True` on a solver has no effect",
)
def test_configure_weight_solver(Simulator, seed, plt, allclose):
    """Ensures that connections that don't use the weight solver ignore it"""
    n1, n2 = 100, 101
    function = lambda x: x ** 2

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Connection].solver = nengo.solvers.LstsqL2(weights=True)

        u = nengo.Node(lambda t: np.sin(8 * t))
        a = nengo.Ensemble(n1, 1)
        b = nengo.Ensemble(n2, 1)
        v = nengo.Node(size_in=1)
        up = nengo.Probe(u, synapse=nengo.Alpha(0.01))
        vp = nengo.Probe(v, synapse=nengo.Alpha(0.01))

        nengo.Connection(u, a)
        ens_conn = nengo.Connection(a, b, function=function)
        nengo.Connection(b, v)

    with nengo.Simulator(net) as sim:
        sim.run(1.0)

    t = sim.trange()
    x = sim.data[up]
    y = function(x)
    z = sim.data[vp]
    assert sim.data[ens_conn].weights.shape == (n2, n1)
    assert signals_allclose(
        t, y, z, buf=0.01, delay=0.015, atol=0.05, rtol=0.05, plt=plt, allclose=allclose
    )


def test_vector(Simulator, AnyNeuronType, plt, seed, allclose):
    N1, N2 = 50, 50
    transform = [-1, 0.5]

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        u = nengo.Node(output=[0.5, 0.5])
        a = nengo.Ensemble(N1, dimensions=2)
        b = nengo.Ensemble(N2, dimensions=2)
        nengo.Connection(u, a)
        nengo.Connection(a, b, transform=transform)

        up = nengo.Probe(u, "output")
        bp = nengo.Probe(b, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.2)
    t = sim.trange()
    x = sim.data[up]
    y = x * transform
    yhat = sim.data[bp]

    plt.plot(t, y, "--")
    plt.plot(t, yhat)

    assert allclose(y[-10:], yhat[-10:], atol=0.1, rtol=0.01)


def test_dimensionality_errors(NonDirectNeuronType, seed, rng):
    N = 10
    with nengo.Network(seed=seed) as m:
        m.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        n01 = nengo.Node(output=[1])
        n02 = nengo.Node(output=[1, 1])
        n21 = nengo.Node(output=lambda t, x: [1], size_in=2)
        e1 = nengo.Ensemble(N, 1)
        e2 = nengo.Ensemble(N, 2)

        # these should work
        nengo.Connection(n01, e1)
        nengo.Connection(n02, e2)
        nengo.Connection(e2, n21)
        nengo.Connection(n21, e1)
        nengo.Connection(e1.neurons, n21, transform=rng.randn(2, N))
        nengo.Connection(e2, e1, function=lambda x: x[0])
        nengo.Connection(e2, e2, transform=np.ones(2))

        # these should not work
        with pytest.raises(
            ValidationError, match="not equal to connection output size"
        ):
            nengo.Connection(n02, e1)
        with pytest.raises(
            ValidationError, match="not equal to connection output size"
        ):
            nengo.Connection(e1, e2)
        with pytest.raises(ValidationError, match="Transform input size"):
            nengo.Connection(
                e2.neurons, e1, transform=Dense((1, N + 1), init=Choice([1.0]))
            )
        with pytest.raises(ValidationError, match="Transform output size"):
            nengo.Connection(
                e2.neurons, e1, transform=Dense((2, N), init=Choice([1.0]))
            )
        with pytest.raises(ValidationError, match="Function output size"):
            nengo.Connection(e2, e1, function=lambda x: x, transform=Dense((1, 1)))
        with pytest.raises(ValidationError, match="function.*must accept a single"):
            nengo.Connection(e2, e1, function=lambda: 0, transform=Dense((1, 1)))
        with pytest.raises(ValidationError, match="Function output size"):
            nengo.Connection(n21, e2, transform=Dense((2, 2)))
        with pytest.raises(ValidationError, match="Shape of initial value"):
            nengo.Connection(e2, e2, transform=np.ones((2, 2, 2)))
        with pytest.raises(ValidationError, match="Function output size"):
            nengo.Connection(e1, e2, transform=Dense((3, 3), init=np.ones(3)))

        # these should not work because of indexing mismatches
        with pytest.raises(ValidationError, match="Function output size"):
            nengo.Connection(n02[0], e2, transform=Dense((2, 2)))
        with pytest.raises(ValidationError, match="Transform output size"):
            nengo.Connection(n02, e2[0], transform=Dense((2, 2)))
        with pytest.raises(ValidationError, match="Function output size"):
            nengo.Connection(n02[1], e2[0], transform=Dense((2, 2)))
        with pytest.raises(ValidationError, match="Transform input size"):
            nengo.Connection(n02, e2[0], transform=Dense((2, 1), init=Choice([1.0])))
        with pytest.raises(ValidationError, match="Transform input size"):
            nengo.Connection(e2[0], e2, transform=Dense((1, 2), init=Choice([1.0])))

        # these should not work because of repeated indices
        dense22 = Dense((2, 2), init=np.ones((2, 2)))
        with pytest.raises(ValidationError, match="Input.*repeated indices"):
            nengo.Connection(n02[[0, 0]], e2, transform=dense22)
        with pytest.raises(ValidationError, match="Output.*repeated indices"):
            nengo.Connection(e2, e2[[1, 1]], transform=dense22)


def test_slicing(Simulator, AnyNeuronType, plt, seed, allclose):
    N = 300

    x = np.array([-1, -0.25, 1])

    s1a = slice(1, None, -1)
    s1b = [2, 0]
    T1 = [[-1, 0.5], [2, 0.25]]
    y1 = np.zeros(3)
    y1[s1b] = np.dot(T1, x[s1a])

    s2a = [0, 2]
    s2b = slice(0, 2)
    T2 = [[-0.5, 0.25], [0.5, 0.75]]
    y2 = np.zeros(3)
    y2[s2b] = np.dot(T2, x[s2a])

    s3a = [2, 0]
    s3b = np.asarray([0, 2])  # test slicing with numpy array
    T3 = [0.5, 0.75]
    y3 = np.zeros(3)
    y3[s3b] = np.dot(np.diag(T3), x[s3a])

    sas = [s1a, s2a, s3a]
    sbs = [s1b, s2b, s3b]
    Ts = [T1, T2, T3]
    ys = [y1, y2, y3]

    weight_solver = nengo.solvers.LstsqL2(weights=True)

    with nengo.Network(seed=seed) as m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()

        u = nengo.Node(output=x)
        a = nengo.Ensemble(N, dimensions=3, radius=1.7)
        nengo.Connection(u, a)

        probes = []
        weight_probes = []
        for sa, sb, T in zip(sas, sbs, Ts):
            b = nengo.Ensemble(N, dimensions=3, radius=1.7)
            nengo.Connection(a[sa], b[sb], transform=T)
            probes.append(nengo.Probe(b, synapse=0.03))

            # also test on weight solver
            b = nengo.Ensemble(N, dimensions=3, radius=1.7)
            nengo.Connection(a[sa], b[sb], transform=T, solver=weight_solver)
            weight_probes.append(nengo.Probe(b, synapse=0.03))

    with Simulator(m) as sim:
        sim.run(0.25)
    t = sim.trange()

    for i, [y, p] in enumerate(zip(ys, probes)):
        plt.subplot(len(ys), 1, i + 1)
        plt.plot(t, np.tile(y, (len(t), 1)), "--")
        plt.plot(t, sim.data[p])

    atol = 0.01 if AnyNeuronType is nengo.Direct else 0.1
    for i, [y, p, wp] in enumerate(zip(ys, probes, weight_probes)):
        assert allclose(y, sim.data[p][-20:], atol=atol), f"Failed {i}"
        assert allclose(y, sim.data[wp][-20:], atol=atol), f"Weights {i}"


def test_neuron_slicing(Simulator, plt, seed, rng, allclose):
    N = 6
    sa = slice(None, None, 2)
    sb = slice(None, None, -2)

    x = np.array([-1, -0.25, 1])
    with nengo.Network(seed=seed) as m:
        m.config[nengo.Ensemble].neuron_type = nengo.LIFRate()

        u = nengo.Node(output=x)
        a = nengo.Ensemble(N, dimensions=3, radius=1.7)
        b = nengo.Ensemble(N, dimensions=3, radius=1.7)
        nengo.Connection(u, a)

        c = nengo.Connection(a.neurons[sa], b.neurons[sb])
        c.transform = rng.normal(scale=1e-3, size=(c.size_out, c.size_in))

        ap = nengo.Probe(a.neurons, synapse=0.03)
        bp = nengo.Probe(b.neurons, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.2)
    t = sim.trange()

    x = sim.data[ap]
    y = np.zeros((len(t), b.n_neurons))
    y[:, sb] = np.dot(x[:, sa], c.transform.init.T)
    y = b.neuron_type.rates(y, sim.data[b].gain, sim.data[b].bias)

    plt.plot(t, y, "k--")
    plt.plot(t, sim.data[bp])
    assert allclose(y[-10:], sim.data[bp][-10:], atol=3.0, rtol=0.0)


def test_shortfilter(Simulator, AnyNeuronType):
    # Testing the case where the connection filter is < dt
    m = nengo.Network()
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        a = nengo.Ensemble(n_neurons=10, dimensions=1)
        nengo.Connection(a, a, synapse=0)

        b = nengo.Ensemble(n_neurons=10, dimensions=1)
        nengo.Connection(a, b, synapse=0)
        nengo.Connection(b, a, synapse=0)

    with Simulator(m, dt=0.01):
        # This test passes if there are no cycles in the op graph
        pass

    # We will still get a cycle if the user explicitly sets the
    # filter to None
    with m:
        d = nengo.Ensemble(1, dimensions=1, neuron_type=nengo.Direct())
        nengo.Connection(d, d, synapse=None)
    with pytest.raises(ValueError):
        Simulator(m, dt=0.01)


def test_zerofilter(Simulator, seed):
    # Testing the case where the connection filter is zero
    m = nengo.Network(seed=seed)
    with m:
        # Ensure no cycles in the op graph.
        a = nengo.Ensemble(1, dimensions=1, neuron_type=nengo.Direct())
        nengo.Connection(a, a, synapse=0)

        # Ensure that spikes are not filtered
        b = nengo.Ensemble(
            3, dimensions=1, intercepts=[-0.9, -0.8, -0.7], neuron_type=nengo.LIF()
        )
        bp = nengo.Probe(b.neurons)

    with Simulator(m) as sim:
        sim.run(1.0)
    # assert that we have spikes (binary)
    assert np.unique(sim.data[bp]).size == 2


def test_function_output_size(Simulator, plt, seed, allclose):
    """Try a function that outputs both 0-d and 1-d arrays"""

    def bad_function(x):
        return x if x > 0 else 0

    model = nengo.Network(seed=seed)
    with model:
        u = nengo.Node(output=lambda t: (t - 0.1) * 10)
        a = nengo.Ensemble(n_neurons=100, dimensions=1)
        b = nengo.Ensemble(n_neurons=100, dimensions=1)
        nengo.Connection(u, a, synapse=None)
        nengo.Connection(a, b, synapse=None, function=bad_function)
        up = nengo.Probe(u)
        bp = nengo.Probe(b, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)
    t = sim.trange()
    x = nengo.Lowpass(0.03).filt(np.maximum(sim.data[up], 0), dt=sim.dt)
    y = sim.data[bp]

    plt.plot(t, x, "k")
    plt.plot(t, y)

    assert allclose(x, y, atol=0.1)


def test_slicing_function(Simulator, plt, seed, allclose):
    """Test using a pre-slice and a function"""
    N = 300
    f_in = lambda t: [np.cos(3 * t), np.sin(3 * t)]
    f_x = lambda x: [x, -(x ** 2)]

    with nengo.Network(seed=seed) as model:
        u = nengo.Node(output=f_in)
        a = nengo.Ensemble(N, 2, radius=np.sqrt(2))
        b = nengo.Ensemble(N, 2, radius=np.sqrt(2))
        nengo.Connection(u, a)
        nengo.Connection(a[1], b, function=f_x)

        up = nengo.Probe(u, synapse=0.05)
        bp = nengo.Probe(b, synapse=0.05)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    v = sim.data[up]
    w = np.column_stack(f_x(v[:, 1]))
    y = sim.data[bp]

    plt.plot(t, y)
    plt.plot(t, w, ":")

    assert allclose(w, y, atol=0.1)


@pytest.mark.parametrize("negative_indices", (True, False))
def test_list_indexing(Simulator, plt, seed, negative_indices, allclose):
    with nengo.Network(seed=seed) as model:
        u = nengo.Node([-1, 1])
        a = nengo.Ensemble(50, dimensions=1)
        b = nengo.Ensemble(50, dimensions=1, radius=2.2)
        c = nengo.Ensemble(100, dimensions=2, radius=1.3)
        d = nengo.Ensemble(100, dimensions=2, radius=1.3)
        if negative_indices:
            nengo.Connection(u[[0, 1]], a[[0, -1]])
            nengo.Connection(u[[1, -1]], b[[0, -1]])
            nengo.Connection(u[[0, 1]], c[[0, 1]])
            nengo.Connection(u[[1, -1]], d[[0, 1]])
        else:
            nengo.Connection(u[[0, 1]], a[[0, 0]])
            nengo.Connection(u[[1, 1]], b[[0, 0]])
            nengo.Connection(u[[0, 1]], c[[0, 1]])
            nengo.Connection(u[[1, 1]], d[[0, 1]])

        a_probe = nengo.Probe(a, synapse=0.03)
        b_probe = nengo.Probe(b, synapse=0.03)
        c_probe = nengo.Probe(c, synapse=0.03)
        d_probe = nengo.Probe(d, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.4)

    t = sim.trange()
    a_data = sim.data[a_probe]
    b_data = sim.data[b_probe]
    c_data = sim.data[c_probe]
    d_data = sim.data[d_probe]

    line = plt.plot(t, a_data)
    plt.axhline(0, color=line[0].get_color())
    line = plt.plot(t, b_data)
    plt.axhline(2, color=line[0].get_color())
    line = plt.plot(t, c_data)
    plt.axhline(-1, color=line[0].get_color())
    line = plt.plot(t, d_data)
    plt.axhline(1, color=line[1].get_color())

    assert allclose(a_data[t > 0.3], [0], atol=0.2)
    assert allclose(b_data[t > 0.3], [2], atol=0.2)
    assert allclose(c_data[t > 0.3], [-1, 1], atol=0.2)
    assert allclose(d_data[t > 0.3], [1, 1], atol=0.2)


@pytest.mark.filterwarnings("ignore:boolean index did not match")
def test_boolean_indexing(Simulator, rng, plt, allclose):
    D = 10
    mu = np.arange(D) % 2 == 0
    mv = np.arange(D) % 2 == 1
    x = rng.uniform(-1, 1, size=D)
    y = np.zeros(D)
    y[mv] = x[mu]

    with nengo.Network() as model:
        u = nengo.Node(x)
        v = nengo.Node(size_in=D)
        nengo.Connection(u[mu], v[mv], synapse=None)
        v_probe = nengo.Probe(v)

    with Simulator(model) as sim:
        sim.run(0.01)

    plt.plot(sim.trange(), sim.data[v_probe])
    assert allclose(sim.data[v_probe][1:], y, atol=1e-5, rtol=1e-3)


def test_set_weight_solver():
    with nengo.Network() as net:
        # set default Gaussian transform to avoid transform errors
        net.config[nengo.Connection].transform = nengo.dists.Gaussian(0, 1)

        ens = nengo.Ensemble(10, 2)
        node = nengo.Node(lambda t, x: x, size_in=2, size_out=2)

        nengo.Connection(ens, ens, solver=LstsqL2(weights=True))

        with pytest.warns(
            UserWarning, match="For connections from.*setting the solver has no effect"
        ):
            nengo.Connection(node, ens, solver=LstsqL2(weights=True))

        with pytest.warns(
            UserWarning, match="For connections to.*setting `weights=True`.*no effect"
        ):
            nengo.Connection(ens, node, solver=LstsqL2(weights=True))


def test_set_learning_rule():
    with nengo.Network():
        a = nengo.Ensemble(10, 2)
        b = nengo.Ensemble(10, 2)
        nengo.Connection(a, b, learning_rule_type=nengo.PES())
        nengo.Connection(
            a, b, learning_rule_type=nengo.PES(), solver=LstsqL2(weights=True)
        )
        nengo.Connection(
            a.neurons, b.neurons, learning_rule_type=nengo.PES(), transform=1
        )
        nengo.Connection(
            a.neurons,
            b.neurons,
            learning_rule_type=nengo.Oja(),
            transform=np.ones((10, 10)),
        )

        n = nengo.Node(output=lambda t, x: t * x, size_in=2)
        with pytest.raises(
            ValidationError, match="'pre' must be of type 'Ensemble'.*PES"
        ):
            nengo.Connection(n, a, learning_rule_type=nengo.PES())


def test_set_function(Simulator):
    with nengo.Network() as model:
        a = nengo.Ensemble(10, 2)
        b = nengo.Ensemble(10, 2)
        c = nengo.Ensemble(10, 1)
        d = nengo.Node(size_in=1)

        # Function only OK from node or ensemble
        with pytest.raises(
            ValidationError, match="function can only be set.*Ensemble or Node"
        ):
            nengo.Connection(a.neurons, b, function=lambda x: x)

        # Function and transform must match up
        with pytest.raises(
            ValidationError,
            match="Shape of initial value.*does not match expected shape",
        ):
            nengo.Connection(a, b, function=lambda x: x[0] * x[1], transform=np.eye(2))

        # No functions allowed on passthrough nodes
        with pytest.raises(
            ValidationError, match="Cannot apply functions to passthrough nodes"
        ):
            nengo.Connection(d, c, function=lambda x: [1])

        # These initial functions have correct dimensionality
        conn_2d = nengo.Connection(a, b)
        conn_1d = nengo.Connection(b, c, function=lambda x: x[0] * x[1])

    with Simulator(model):
        pass  # Builds fine

    with model:
        # Can change to another function with correct dimensionality
        conn_2d.function = lambda x: x ** 2
        conn_1d.function = lambda x: x[0] + x[1]

    with Simulator(model):
        pass  # Builds fine


def test_set_eval_points(Simulator):
    with nengo.Network() as model:
        a = nengo.Ensemble(
            10,
            2,
            encoders=nengo.dists.Choice([[1, 1]]),
            intercepts=nengo.dists.Uniform(-1, -0.1),
        )
        b = nengo.Ensemble(10, 2)
        # ConnEvalPoints parameter checks that pre is an ensemble
        nengo.Connection(a, b, eval_points=[[0, 0], [0.5, 1]])
        nengo.Connection(a, b, eval_points=nengo.dists.Uniform(0, 1))
        with pytest.raises(
            ValidationError,
            match="eval_points are only valid on connections from ensembles",
        ):
            nengo.Connection(a.neurons, b, eval_points=[[0, 0], [0.5, 1]])

    with Simulator(model):
        pass  # Builds fine


@pytest.mark.parametrize("sample", [False, True])
@pytest.mark.parametrize("radius", [0.5, 1.0, 1.5])
@pytest.mark.parametrize("scale", [False, True])
def test_eval_points_scaling(Simulator, sample, radius, seed, rng, scale):
    eval_points = UniformHypersphere()
    if sample:
        eval_points = eval_points.sample(500, 3, rng=rng)

    model = nengo.Network(seed=seed)
    with model:
        a = nengo.Ensemble(
            1,
            3,
            encoders=np.ones((1, 3)),
            intercepts=nengo.dists.Choice([-1]),
            radius=radius,
        )
        b = nengo.Ensemble(1, 3)
        con = nengo.Connection(a, b, eval_points=eval_points, scale_eval_points=scale)

    with Simulator(model) as sim:
        dists = npext.norm(sim.data[con].eval_points, axis=1)
    limit = radius if scale else 1.0
    assert np.all(dists <= limit)
    assert np.any(dists >= 0.9 * limit)


def test_solverparam():
    """SolverParam must be a solver."""

    class Test:
        sp = ConnectionSolverParam("sp", default=None)

    inst = Test()
    assert inst.sp is None
    inst.sp = LstsqL2()
    assert isinstance(inst.sp, LstsqL2)
    assert not inst.sp.weights
    # Non-solver not OK
    with pytest.raises(ValidationError):
        inst.sp = "a"


def test_prepost_errors(Simulator):
    with nengo.Network():
        x = nengo.Ensemble(10, 1)
        y = nengo.Ensemble(10, 1)
        c = nengo.Connection(x, y)
        with pytest.raises(ValueError):
            nengo.Connection(3, x)  # from non-NengoObject
        with pytest.raises(ValueError):
            nengo.Connection(x, 3)  # to non-NengoObject
        with pytest.raises(ValueError):
            nengo.Connection(c, x)  # from connection
        with pytest.raises(ValueError):
            nengo.Connection(x, c)  # to connection

    # --- non-existent objects
    with nengo.Network():
        a = nengo.Ensemble(10, 1)

    # non-existent pre
    with nengo.Network() as model1:
        e1 = nengo.Ensemble(10, 1)
        nengo.Connection(a, e1)
    with pytest.raises(ValueError):
        Simulator(model1)

    # non-existent post
    with nengo.Network() as model2:
        e2 = nengo.Ensemble(10, 1)
        nengo.Connection(e2, a)
    with pytest.raises(ValueError):
        Simulator(model2)

    # probe non-existent object
    with nengo.Network() as model3:
        nengo.Probe(a)
    with pytest.raises(ValueError):
        Simulator(model3)


def test_directneurons(NonDirectNeuronType):
    with nengo.Network():
        a = nengo.Ensemble(1, 1, neuron_type=NonDirectNeuronType())
        b = nengo.Ensemble(1, 1, neuron_type=nengo.Direct())

        # cannot connect to or from direct neurons
        with pytest.raises(ValidationError, match="must have size_in > 0"):
            nengo.Connection(a, b.neurons)
        with pytest.raises(ValidationError, match="must have size_out > 0"):
            nengo.Connection(b.neurons, a)


def test_decoder_probe(Simulator):
    """Ensures we can only probe decoders in connections from ensembles."""
    with nengo.Network() as net:
        pre = nengo.Ensemble(10, 10)
        post = nengo.Ensemble(10, 10)
        c_ens = nengo.Connection(pre, post)
        c_ens_neurons = nengo.Connection(pre, post.neurons)
        nengo.Probe(c_ens, "weights")
        nengo.Probe(c_ens_neurons, "weights")

    with Simulator(net) as sim:
        assert sim


def test_transform_probe(Simulator, allclose):
    """Ensures we can always probe transform in connections."""
    with nengo.Network() as net:
        pre = nengo.Ensemble(10, 10)
        post = nengo.Ensemble(10, 10)

        c_ens = nengo.Connection(
            pre, post, solver=nengo.solvers.NoSolver(np.ones((10, 10)))
        )
        c_ens_neurons = nengo.Connection(
            pre, post.neurons, solver=nengo.solvers.NoSolver(np.ones((10, 10)))
        )
        c_neurons_ens = nengo.Connection(pre.neurons, post, transform=1)
        c_neurons = nengo.Connection(pre.neurons, post.neurons, transform=1)

        p_neurons_ens = nengo.Probe(c_neurons_ens, "weights")
        p_neurons = nengo.Probe(c_neurons, "weights")
        p_ens = nengo.Probe(c_ens, "weights")
        p_ens_neurons = nengo.Probe(c_ens_neurons, "weights")

    with Simulator(net) as sim:
        sim.step()
        assert allclose(sim.data[p_neurons_ens], 1)
        assert allclose(sim.data[p_neurons], 1)
        assert allclose(sim.data[p_ens], 1)
        assert allclose(sim.data[p_ens_neurons], 1)

    with net:
        c_neurons_ens_none = nengo.Connection(pre.neurons, post, transform=None)
        nengo.Probe(c_neurons_ens_none, "weights")

    with pytest.raises(BuildError, match="cannot be probed"):
        Simulator(net)


def test_connectionlearningruletypeparam():
    with nengo.Network():
        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(11, 1)

        with pytest.raises(
            ValidationError, match="can only be applied on connections to neurons"
        ):
            nengo.Connection(a, b, learning_rule_type=nengo.BCM())

        with pytest.raises(ValidationError, match="does not match expected shape"):
            nengo.Connection(
                a, b, transform=np.ones((10, 11)), learning_rule_type=nengo.BCM()
            )


def test_function_with_no_name(Simulator):
    def add(x, val):
        return x + val

    model = nengo.Network()

    with model:
        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(11, 1)
        nengo.Connection(a, b, function=partial(add, val=2))

    with Simulator(model) as sim:
        assert sim


def test_function_points(Simulator, seed, rng, plt, allclose):
    x = rng.uniform(-1, 1, size=(1000, 1))
    y = -x

    with nengo.Network(seed=seed) as model:
        u = nengo.Node(nengo.processes.WhiteSignal(1.0, high=9, rms=0.3))
        a = nengo.Ensemble(100, 1)
        v = nengo.Node(size_in=1)
        nengo.Connection(u, a, synapse=None)
        nengo.Connection(a, v, eval_points=x, function=y)

        up = nengo.Probe(u, synapse=nengo.Alpha(0.01))
        vp = nengo.Probe(v, synapse=nengo.Alpha(0.01))

    with Simulator(model, seed=seed) as sim:
        sim.run(1.0)

    assert signals_allclose(
        sim.trange(),
        -sim.data[up],
        sim.data[vp],
        buf=0.01,
        delay=0.005,
        atol=5e-2,
        rtol=3e-2,
        plt=plt,
        allclose=allclose,
    )


def test_connectionfunctionparam_array(Simulator, seed):
    points_1d = np.zeros((100, 1))
    points_2d = np.zeros((100, 2))
    points_v = np.zeros((100,))
    points_50 = np.zeros((50, 1))

    with nengo.Network(seed=seed) as model:
        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(10, 1)

        # eval points not set raises error
        with pytest.raises(ValidationError, match="'eval_points' must also be set"):
            nengo.Connection(a, b, function=points_1d)

        # wrong ndims raises error
        with pytest.raises(ValidationError, match="array must be 2D"):
            nengo.Connection(a, b, eval_points=points_1d, function=points_v)

        # wrong number of points raises error
        with pytest.raises(
            ValidationError,
            match="Number of eval.*must match number of function points",
        ):
            nengo.Connection(a, b, eval_points=points_50, function=points_1d)

        # wrong output dims raises error
        with pytest.raises(
            ValidationError,
            match="Transform output size.*not equal to connection output",
        ):
            nengo.Connection(a, b, eval_points=points_1d, function=points_2d)

        nengo.Connection(a, b, eval_points=points_1d, function=points_1d)
        nengo.Connection(
            a, b, eval_points=points_1d, function=points_2d, transform=np.ones((1, 2))
        )

    with Simulator(model):
        pass


def test_function_names():
    def my_func(x):
        return x

    class MyFunc:
        def __call__(self, x):
            return x

    with nengo.Network():
        a = nengo.Ensemble(10, 1)
        b = nengo.Node(size_in=1)

        func_conn = nengo.Connection(a, b, function=my_func)
        assert str(func_conn).endswith("computing 'my_func'>")

        class_conn = nengo.Connection(a, b, function=MyFunc())
        assert str(class_conn).endswith("computing 'MyFunc'>")

        array_conn = nengo.Connection(
            a, b, eval_points=np.zeros((10, 1)), function=np.zeros((10, 1))
        )
        assert str(array_conn).endswith("computing 'ndarray'>")


@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value")
@pytest.mark.filterwarnings("ignore:Non-finite values detected")
def test_zero_activities_error(Simulator):
    with nengo.Network() as model:
        a = nengo.Ensemble(10, 1)
        a.gain = np.zeros(10)
        a.bias = np.zeros(10)
        nengo.Connection(a, nengo.Node(size_in=1))

    with pytest.raises(BuildError, match="'activities' matrix is all zero"):
        with Simulator(model):
            pass


def test_function_returns_none_error(Simulator):
    with nengo.Network() as model:
        a = nengo.Ensemble(10, 1)
        nengo.Connection(
            a,
            nengo.Node(size_in=1),
            eval_points=[[0], [1]],
            function=lambda x: x if x < 0.5 else None,
        )

    with pytest.raises(BuildError):
        with Simulator(model):
            pass


def test_transform_none():
    with nengo.Network():
        a = nengo.Node([0])
        b = nengo.Node(size_in=1)
        conn = nengo.Connection(a, b, transform=None)
        assert isinstance(conn.transform, NoTransform)


@pytest.mark.filterwarnings("ignore:divide by zero")
def test_neuron_advanced_indexing(Simulator):
    with nengo.Network() as net:
        node = nengo.Node([1, 1, 1])
        ens = nengo.Ensemble(
            5, 1, bias=nengo.dists.Choice([0]), gain=nengo.dists.Choice([1])
        )
        nengo.Connection(node, ens.neurons[[0, 0, 2]], synapse=None)
        p = nengo.Probe(ens.neurons, "input")

    with Simulator(net) as sim:
        sim.run(0.001)
        assert np.allclose(sim.data[p], [2, 0, 1, 0, 0])


def test_learning_rule_equality():
    with nengo.Network():
        ens = nengo.Ensemble(10, 1)
        lr = nengo.PES()
        conn0 = nengo.Connection(ens, ens, learning_rule_type=lr)
        assert conn0.learning_rule == conn0.learning_rule
        assert conn0.learning_rule != lr

        conn1 = nengo.Connection(ens, ens, learning_rule_type=[lr, nengo.PES()])
        assert conn0.learning_rule[0] != conn1.learning_rule
        assert conn1.learning_rule_type[0] == conn0.learning_rule_type
        assert conn1.learning_rule[0] == conn1.learning_rule[1]


def test_learning_transform_shape_error(Simulator):
    with nengo.Network() as net:
        a = nengo.Ensemble(10, dimensions=2)
        b = nengo.Ensemble(10, dimensions=2)
        nengo.Connection(
            a.neurons, b.neurons, transform=1, learning_rule_type=nengo.BCM()
        )

    with pytest.raises(
        BuildError, match="'transform' must be a 2-dimensional array for learning"
    ):
        with Simulator(net):
            pass


def test_is_decoded_deprecation():
    with pytest.warns(DeprecationWarning, match="is_decoded is deprecated"):
        with nengo.Network():
            assert nengo.Connection(nengo.Node(0), nengo.Node(size_in=1)).is_decoded


def test_bad_function_type():
    with nengo.Network():
        ens = nengo.Ensemble(10, 1)
        with pytest.raises(ValidationError, match="Invalid connection function type"):
            nengo.Connection(ens, ens, function="hi")


def test_probeable():
    with nengo.Network():
        conn = nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))
        assert conn.probeable == ("output", "input", "weights")



from fnmatch import fnmatch
from inspect import getfullargspec

import numpy as np
import pytest

import nengo
from nengo import (
    BCM,
    LIF,
    PES,
    RLS,
    AdaptiveLIF,
    AdaptiveLIFRate,
    Alpha,
    Conv,
    Convolution,
    ConvTranspose,
    Dense,
    Direct,
    Izhikevich,
    LIFRate,
    LinearFilter,
    Lowpass,
    Oja,
    RectifiedLinear,
    Sigmoid,
    Sparse,
    SpikingRectifiedLinear,
    Tanh,
    Voja,
)
from nengo.builder.learning_rules import SimBCM, SimOja, SimPES, SimRLS, SimVoja
from nengo.builder.neurons import SimNeurons
from nengo.builder.operator import (
    Copy,
    DotInc,
    ElementwiseInc,
    Reset,
    SimPyFunc,
    TimeUpdate,
)
from nengo.builder.processes import SimProcess
from nengo.builder.signal import Signal
from nengo.dists import (
    PDF,
    Choice,
    CosineSimilarity,
    Exponential,
    Gaussian,
    Samples,
    ScatteredHypersphere,
    SqrtBeta,
    SubvectorLength,
    Uniform,
    UniformHypersphere,
)
from nengo.processes import (
    BrownNoise,
    FilteredNoise,
    Piecewise,
    PresentInput,
    WhiteNoise,
    WhiteSignal,
)
from nengo.solvers import (
    Lstsq,
    LstsqL1,
    LstsqL2,
    LstsqL2nz,
    LstsqNoise,
    Nnls,
    NnlsL2,
    NnlsL2nz,
    NoSolver,
)
from nengo.synapses import Triangle
from nengo.transforms import (
    ChannelShape,
    ConvolutionTranspose,
    NoTransform,
    SparseMatrix,
)


def check_init_args(cls, args):
    assert getfullargspec(cls.__init__).args[1:] == args


def check_repr(obj):
    """Tests that repr gives back what is needed to create the object."""
    assert not isinstance(obj, str), "Passing a string doesn't test `repr(obj)`"

    # some reprs need these in the local namespace
    array = np.array
    int64 = np.int64
    assert array
    assert int64
    assert eval(repr(obj)) == obj  # pylint: disable = eval-used


def test_core_objects():
    with nengo.Network() as net1:

        a = nengo.Ensemble(100, 2)
        with nengo.Network(label="net2") as net2:
            b = nengo.Ensemble(100, 2, label="b")

        ap = nengo.Probe(a)
        bp = nengo.Probe(b)

        c1 = nengo.Connection(a, b)
        c2 = nengo.Connection(a, b, function=np.square)
        c3 = nengo.Connection(a, b, learning_rule_type=nengo.PES(), label="c3")

        n1 = nengo.Node(output=np.sin)
        n2 = nengo.Node(output=np.cos, label="n2")

    assert fnmatch(repr(net1), "<Network (unlabeled) at 0x*>")
    assert fnmatch(repr(net2), '<Network "net2" at 0x*>')

    # Ensemble
    assert fnmatch(repr(a), "<Ensemble (unlabeled) at 0x*>")
    assert fnmatch(repr(b), "<Ensemble 'b' at 0x*>")

    # Probe
    assert fnmatch(
        repr(ap), "<Probe at 0x* of 'decoded_output' of <Ensemble (unlabeled) at 0x*>>"
    )
    assert fnmatch(repr(bp), "<Probe at 0x* of 'decoded_output' of <Ensemble 'b'>>")

    # Connection
    assert fnmatch(
        repr(c1),
        "<Connection at 0x* from <Ensemble (unlabeled) at 0x*> to <Ensemble 'b'>>",
    )
    assert fnmatch(
        repr(c2),
        "<Connection at 0x* from <Ensemble (unlabeled) at 0x*> to <Ensemble 'b'> "
        "computing 'square'>",
    )
    assert fnmatch(repr(c3), "<Connection at 0x* c3>")

    assert fnmatch(
        repr(c3.learning_rule),
        "<LearningRule at 0x* modifying <Connection at 0x* c3> with type PES()>",
    )

    # Node
    assert fnmatch(repr(n1), "<Node (unlabeled) at 0x*>")
    assert fnmatch(repr(n2), "<Node 'n2' at 0x*>")

    # Neurons
    assert fnmatch(repr(b.neurons), "<Neurons at 0x* of <Ensemble 'b' at 0x*>>")

    # ObjView
    assert fnmatch(repr(a[:1]), "<Ensemble (unlabeled) at 0x*>[[]:1[]]")


def test_neuron_types():
    check_init_args(Direct, ["initial_state"])
    check_repr(Direct())
    assert repr(Direct()) == "Direct()"

    check_init_args(RectifiedLinear, ["amplitude", "initial_state"])
    check_repr(RectifiedLinear())
    check_repr(RectifiedLinear(amplitude=2))
    assert repr(RectifiedLinear()) == "RectifiedLinear()"

    check_init_args(SpikingRectifiedLinear, ["amplitude", "initial_state"])
    check_repr(SpikingRectifiedLinear())
    check_repr(SpikingRectifiedLinear(amplitude=2))
    check_repr(SpikingRectifiedLinear(initial_state={"voltage": Choice([1.0])}))
    assert repr(SpikingRectifiedLinear()) == "SpikingRectifiedLinear()"

    check_init_args(Sigmoid, ["tau_ref", "initial_state"])
    check_repr(Sigmoid())
    check_repr(Sigmoid(tau_ref=0.1))
    assert repr(Sigmoid()), "Sigmoid()"
    assert repr(Sigmoid(tau_ref=0.001)) == "Sigmoid(tau_ref=0.001)"

    check_init_args(LIFRate, ["tau_rc", "tau_ref", "amplitude", "initial_state"])
    check_repr(LIFRate())
    check_repr(LIFRate(tau_rc=0.1))
    check_repr(LIFRate(tau_ref=0.1))
    check_repr(LIFRate(amplitude=2))
    check_repr(LIFRate(tau_rc=0.05, tau_ref=0.02))
    check_repr(LIFRate(tau_rc=0.05, amplitude=2))
    check_repr(LIFRate(tau_ref=0.02, amplitude=2))
    check_repr(LIFRate(tau_rc=0.05, tau_ref=0.02, amplitude=2))
    assert repr(LIFRate()) == "LIFRate()"
    assert repr(LIFRate(tau_rc=0.01, tau_ref=0)) == "LIFRate(tau_rc=0.01, tau_ref=0)"

    check_init_args(
        LIF, ["tau_rc", "tau_ref", "min_voltage", "amplitude", "initial_state"]
    )
    check_repr(LIF())
    check_repr(LIF(tau_rc=0.1))
    check_repr(LIF(tau_ref=0.1))
    check_repr(LIF(amplitude=2))
    check_repr(LIF(min_voltage=-0.5))
    check_repr(LIF(tau_rc=0.05, tau_ref=0.02))
    check_repr(LIF(tau_rc=0.05, amplitude=2))
    check_repr(LIF(tau_ref=0.02, amplitude=2))
    check_repr(LIF(tau_rc=0.05, tau_ref=0.02, amplitude=2))
    check_repr(LIF(tau_rc=0.05, tau_ref=0.02, min_voltage=-0.5, amplitude=2))
    check_repr(LIF(initial_state={"refractory_time": Choice([0.1])}))
    assert repr(LIF()) == "LIF()"
    assert repr(LIF(tau_rc=0.01, tau_ref=0)) == "LIF(tau_rc=0.01, tau_ref=0)"

    check_init_args(
        AdaptiveLIFRate,
        ["tau_n", "inc_n", "tau_rc", "tau_ref", "amplitude", "initial_state"],
    )
    check_repr(AdaptiveLIFRate())
    check_repr(AdaptiveLIFRate(tau_n=0.1))
    check_repr(AdaptiveLIFRate(inc_n=0.5))
    check_repr(AdaptiveLIFRate(tau_rc=0.1))
    check_repr(AdaptiveLIFRate(tau_ref=0.1))
    check_repr(AdaptiveLIFRate(amplitude=2))
    check_repr(
        AdaptiveLIFRate(tau_n=0.1, inc_n=0.5, tau_rc=0.05, tau_ref=0.02, amplitude=2)
    )
    check_repr(AdaptiveLIFRate(initial_state={"adaptation": Choice([0.1])}))
    assert repr(AdaptiveLIFRate()) == "AdaptiveLIFRate()"
    assert (
        repr(AdaptiveLIFRate(tau_rc=0.01, tau_n=0.5, inc_n=0.02))
        == "AdaptiveLIFRate(tau_n=0.5, inc_n=0.02, tau_rc=0.01)"
    )

    check_init_args(
        AdaptiveLIF,
        [
            "tau_n",
            "inc_n",
            "tau_rc",
            "tau_ref",
            "min_voltage",
            "amplitude",
            "initial_state",
        ],
    )
    check_repr(AdaptiveLIF())
    check_repr(AdaptiveLIF(tau_n=0.1))
    check_repr(AdaptiveLIF(inc_n=0.5))
    check_repr(AdaptiveLIF(tau_rc=0.1))
    check_repr(AdaptiveLIF(tau_ref=0.1))
    check_repr(AdaptiveLIF(min_voltage=-0.5))
    check_repr(
        AdaptiveLIF(
            tau_n=0.1,
            inc_n=0.5,
            tau_rc=0.05,
            tau_ref=0.02,
            min_voltage=-0.5,
            amplitude=2,
        )
    )
    check_repr(AdaptiveLIF(initial_state={"adaptation": Choice([0.1])}))
    assert repr(AdaptiveLIF()) == "AdaptiveLIF()"
    assert (
        repr(AdaptiveLIF(tau_rc=0.01, tau_n=0.5, inc_n=0.02))
        == "AdaptiveLIF(tau_n=0.5, inc_n=0.02, tau_rc=0.01)"
    )

    check_init_args(
        Izhikevich,
        [
            "tau_recovery",
            "coupling",
            "reset_voltage",
            "reset_recovery",
            "initial_state",
        ],
    )
    check_repr(Izhikevich())
    check_repr(Izhikevich(tau_recovery=0.1))
    check_repr(Izhikevich(coupling=0.3))
    check_repr(Izhikevich(reset_voltage=-1))
    check_repr(Izhikevich(reset_recovery=5))
    check_repr(
        Izhikevich(tau_recovery=0.1, coupling=0.3, reset_voltage=-1, reset_recovery=5)
    )
    check_repr(Izhikevich(initial_state={"recovery": Choice([0.1])}))
    assert repr(Izhikevich()) == "Izhikevich()"
    assert (
        repr(
            Izhikevich(
                tau_recovery=0.01, coupling=0.5, reset_voltage=-60, reset_recovery=6
            )
        )
        == "Izhikevich(tau_recovery=0.01, coupling=0.5, reset_voltage=-60, "
        "reset_recovery=6)"
    )

    check_init_args(Tanh, ["tau_ref", "initial_state"])
    check_repr(Tanh())
    check_repr(Tanh(tau_ref=0.1))


def test_learning_rule_types():
    check_init_args(PES, ["learning_rate", "pre_synapse"])
    check_repr(PES(learning_rate=0.1, pre_synapse=Lowpass(tau=0.2)))
    assert repr(PES()) == "PES()"
    assert (
        repr(PES(learning_rate=0.1, pre_synapse=0.2))
        == "PES(learning_rate=0.1, pre_synapse=Lowpass(tau=0.2))"
    )

    check_init_args(
        BCM, ["learning_rate", "pre_synapse", "post_synapse", "theta_synapse"]
    )
    check_repr(
        BCM(learning_rate=0.1, pre_synapse=0.2, post_synapse=0.3, theta_synapse=0.4)
    )
    assert repr(BCM()) == "BCM()"
    assert repr(
        BCM(learning_rate=0.1, pre_synapse=0.2, post_synapse=0.3, theta_synapse=0.4)
    ) == (
        "BCM(learning_rate=0.1, pre_synapse=Lowpass(tau=0.2), "
        "post_synapse=Lowpass(tau=0.3), theta_synapse=Lowpass(tau=0.4))"
    )

    check_init_args(Oja, ["learning_rate", "pre_synapse", "post_synapse", "beta"])
    check_repr(
        Oja(
            learning_rate=0.1,
            pre_synapse=Lowpass(tau=0.2),
            post_synapse=Lowpass(tau=0.3),
            beta=0.4,
        )
    )
    assert repr(Oja()) == "Oja()"
    assert repr(
        Oja(learning_rate=0.1, pre_synapse=0.2, post_synapse=0.3, beta=0.4)
    ) == (
        "Oja(learning_rate=0.1, pre_synapse=Lowpass(tau=0.2), "
        "post_synapse=Lowpass(tau=0.3), beta=0.4)"
    )

    check_init_args(Voja, ["learning_rate", "post_synapse"])
    check_repr(Voja(learning_rate=0.1, post_synapse=Lowpass(tau=0.2)))
    assert repr(Voja()) == "Voja()"
    assert (
        repr(Voja(learning_rate=0.1, post_synapse=0.2))
        == "Voja(learning_rate=0.1, post_synapse=Lowpass(tau=0.2))"
    )

    check_init_args(RLS, ["learning_rate", "pre_synapse"])
    check_repr(RLS(2.4e-3))
    check_repr(RLS(learning_rate=0.1, pre_synapse=Alpha(tau=0.2)))
    assert repr(RLS()) == "RLS()"
    assert (
        repr(RLS(learning_rate=0.1, pre_synapse=0.2))
        == "RLS(learning_rate=0.1, pre_synapse=Lowpass(tau=0.2))"
    )


def test_distributions():
    check_init_args(PDF, ["x", "p"])
    check_repr(PDF([1, 2, 3], [0.1, 0.8, 0.1]))
    assert (
        repr(PDF([1, 2], [0.4, 0.6])) == "PDF(x=array([1., 2.]), p=array([0.4, 0.6]))"
    )

    check_init_args(Uniform, ["low", "high", "integer"])
    check_repr(Uniform(1, 3))
    check_repr(Uniform(1, 4, integer=True))
    assert repr(Uniform(0, 1)) == "Uniform(low=0, high=1)"
    assert repr(Uniform(0, 5, integer=True)) == "Uniform(low=0, high=5, integer=True)"

    check_init_args(Gaussian, ["mean", "std"])
    check_repr(Gaussian(0, 2))
    assert repr(Gaussian(1, 0.1)) == "Gaussian(mean=1, std=0.1)"

    check_init_args(Exponential, ["scale", "shift", "high"])
    check_repr(Exponential(2.0))
    check_repr(Exponential(2.0, shift=0.1))
    check_repr(Exponential(2.0, shift=0.1, high=10.0))
    assert repr(Exponential(2.0)) == "Exponential(scale=2.0)"

    check_init_args(UniformHypersphere, ["surface", "min_magnitude"])
    check_repr(UniformHypersphere())
    check_repr(UniformHypersphere(surface=True))
    check_repr(UniformHypersphere(min_magnitude=0.3))
    assert repr(UniformHypersphere()) == "UniformHypersphere()"
    assert repr(UniformHypersphere(surface=True)) == "UniformHypersphere(surface=True)"

    check_init_args(
        ScatteredHypersphere, ["surface", "min_magnitude", "base", "method"]
    )
    check_repr(ScatteredHypersphere())
    check_repr(ScatteredHypersphere(surface=True))
    check_repr(ScatteredHypersphere(min_magnitude=0.3))
    check_repr(ScatteredHypersphere(base=Uniform(0, 1)))
    check_repr(ScatteredHypersphere(method="tfww"))
    assert repr(ScatteredHypersphere()) == "ScatteredHypersphere()"
    assert (
        repr(ScatteredHypersphere(surface=True)) == "ScatteredHypersphere(surface=True)"
    )
    assert (
        repr(ScatteredHypersphere(base=Uniform(0, 1), method="tfww"))
        == "ScatteredHypersphere(base=Uniform(low=0, high=1), method='tfww')"
    )

    check_init_args(Choice, ["options", "weights"])
    check_repr(Choice([3, 2, 1]))
    check_repr(Choice([3, 2, 1], weights=[0.1, 0.2, 0.7]))
    assert repr(Choice([1, 2, 3])) == "Choice(options=array([1., 2., 3.]))"
    assert (
        repr(Choice([1, 2, 3], weights=[0.1, 0.5, 0.4]))
        == "Choice(options=array([1., 2., 3.]), weights=array([0.1, 0.5, 0.4]))"
    )

    check_init_args(Samples, ["samples"])
    check_repr(Samples([3, 2, 1]))
    assert repr(Samples([3, 2, 1])) == "Samples(samples=array([3., 2., 1.]))"

    check_init_args(SqrtBeta, ["n", "m"])
    check_repr(SqrtBeta(3))
    check_repr(SqrtBeta(3, m=2))
    assert repr(SqrtBeta(3)) == "SqrtBeta(n=3)"
    assert repr(SqrtBeta(3, 2)) == "SqrtBeta(n=3, m=2)"

    check_init_args(SubvectorLength, ["dimensions", "subdimensions"])
    check_repr(SubvectorLength(6))
    check_repr(SubvectorLength(6, 2))
    assert repr(SubvectorLength(3)) == "SubvectorLength(dimensions=3)"

    check_init_args(CosineSimilarity, ["dimensions"])
    check_repr(CosineSimilarity(6))
    assert repr(CosineSimilarity(6)) == "CosineSimilarity(dimensions=6)"


def test_synapses():
    check_init_args(LinearFilter, ["num", "den", "analog", "method"])
    check_repr(LinearFilter([1, 2], [3, 4]))
    check_repr(LinearFilter([1, 2], [3, 4], analog=False))
    assert (
        repr(LinearFilter([1], [0.03, 1]))
        == "LinearFilter(num=array([1.]), den=array([0.03, 1.  ]))"
    )

    check_init_args(Lowpass, ["tau"])
    check_repr(Lowpass(0.3))
    assert repr(Lowpass(0.01)) == "Lowpass(tau=0.01)"

    check_init_args(Alpha, ["tau"])
    check_repr(Alpha(0.3))
    assert repr(Alpha(0.02)) == "Alpha(tau=0.02)"

    check_init_args(Triangle, ["t"])
    check_repr(Triangle(0.3))
    assert repr(Triangle(0.03)) == "Triangle(t=0.03)"


def test_processes():
    assert (
        repr(WhiteSignal(0.2, 10, rms=0.3))
        == "WhiteSignal(period=0.2, high=10, rms=0.3)"
    )

    check_init_args(WhiteNoise, ["dist", "scale"])
    check_repr(WhiteNoise(scale=False))
    assert repr(WhiteNoise()) == "WhiteNoise()"
    assert repr(WhiteNoise(scale=False)) == "WhiteNoise(scale=False)"

    check_init_args(FilteredNoise, ["synapse", "dist", "scale"])
    check_repr(FilteredNoise(scale=False))
    assert repr(FilteredNoise()) == "FilteredNoise()"
    assert repr(FilteredNoise(scale=False)) == "FilteredNoise(scale=False)"

    check_init_args(BrownNoise, ["dist"])
    check_repr(BrownNoise())
    assert repr(BrownNoise()) == "BrownNoise()"

    check_init_args(PresentInput, ["inputs", "presentation_time"])
    check_repr(PresentInput(inputs=np.array([1.2, 3.4]), presentation_time=5))
    assert (
        repr(PresentInput((1.2, 3.4), 5))
        == "PresentInput(inputs=array([1.2, 3.4]), presentation_time=5)"
    )

    check_init_args(WhiteSignal, ["period", "high", "rms", "y0"])
    check_repr(WhiteSignal(period=1.2, high=3.4, rms=5.6, y0=7.8))
    assert repr(WhiteSignal(1, 2)) == "WhiteSignal(period=1, high=2)"
    assert (
        repr(WhiteSignal(1.2, 3.4, 5.6, 7.8))
        == "WhiteSignal(period=1.2, high=3.4, rms=5.6, y0=7.8)"
    )

    check_init_args(Piecewise, ["data", "interpolation"])
    check_repr(Piecewise(data={1: 0.1, 2: 0.2, 3: 0.3}))
    assert (
        repr(Piecewise({1: 0.1, 2: 0.2, 3: 0.3}))
        == "Piecewise(data={1: array([0.1]), 2: array([0.2]), 3: array([0.3])})"
    )


def test_piecewise():
    pytest.importorskip("scipy.optimize")
    for interpolation in ("linear", "nearest", "slinear", "quadratic", "cubic"):
        assert repr(Piecewise({1: 0.1, 2: 0.2, 3: 0.3}, interpolation)) == (
            "Piecewise(data={1: array([0.1]), 2: array([0.2]), 3: array([0.3])}, "
            f"interpolation='{interpolation}')"
        )


def test_solvers():
    check_init_args(Lstsq, ["weights", "rcond"])
    check_repr(Lstsq(weights=True, rcond=0.1))
    assert repr(Lstsq(weights=True, rcond=0.1)) == "Lstsq(weights=True, rcond=0.1)"

    check_init_args(LstsqNoise, ["weights", "noise", "solver"])
    check_repr(LstsqNoise(weights=True, noise=0.2))
    assert (
        repr(LstsqNoise(weights=True, noise=0.2))
        == "LstsqNoise(weights=True, noise=0.2)"
    )

    check_init_args(LstsqL2, ["weights", "reg", "solver"])
    check_repr(LstsqL2(weights=True, reg=0.2))
    assert repr(LstsqL2(weights=True, reg=0.2)) == "LstsqL2(weights=True, reg=0.2)"

    check_init_args(LstsqL2nz, ["weights", "reg", "solver"])
    check_repr(LstsqL2nz(weights=True, reg=0.2))
    assert repr(LstsqL2nz(weights=True, reg=0.2)) == "LstsqL2nz(weights=True, reg=0.2)"

    check_init_args(NoSolver, ["values", "weights"])
    check_repr(NoSolver(values=np.array([[1.2, 3.4, 5.6, 7.8]]), weights=True))
    assert (
        repr(NoSolver([[1.2, 3.4, 5.6, 7.8]], weights=True))
        == "NoSolver(values=array([[1.2, 3.4, 5.6, 7.8]]), weights=True)"
    )


def test_lstsql1_repr():
    pytest.importorskip("sklearn")

    check_init_args(LstsqL1, ["weights", "l1", "l2", "max_iter"])
    check_repr(LstsqL1(weights=True, l1=0.2, l2=0.3, max_iter=4))
    assert (
        repr(LstsqL1(weights=True, l1=0.2, l2=0.3, max_iter=4))
        == "LstsqL1(weights=True, l1=0.2, l2=0.3, max_iter=4)"
    )


def test_nnls_repr():
    pytest.importorskip("scipy.optimize")

    check_init_args(Nnls, ["weights"])
    check_repr(Nnls(weights=True))
    assert repr(Nnls(weights=True)) == "Nnls(weights=True)"

    check_init_args(NnlsL2, ["weights", "reg"])
    check_repr(NnlsL2(weights=True, reg=0.2))
    assert repr(NnlsL2(weights=True, reg=0.2)) == "NnlsL2(weights=True, reg=0.2)"

    check_init_args(NnlsL2nz, ["weights", "reg"])
    check_repr(NnlsL2nz(weights=True, reg=0.2))
    assert repr(NnlsL2nz(weights=True, reg=0.2)) == "NnlsL2nz(weights=True, reg=0.2)"


def test_transforms():
    check_init_args(Dense, ["shape", "init"])
    # No check_repr because dense matrices are usually too big
    assert repr(Dense((1, 2), init=[[1, 1]])) == "Dense(shape=(1, 2))"

    check_init_args(
        Conv,
        [
            "n_filters",
            "input_shape",
            "kernel_size",
            "strides",
            "padding",
            "channels_last",
            "init",
        ],
    )
    check_repr(Convolution(n_filters=3, input_shape=(3, 4, 5)))
    check_repr(Convolution(n_filters=3, input_shape=(3, 2, 5), kernel_size=(3, 2)))
    check_repr(Convolution(n_filters=3, input_shape=(3, 4, 5), channels_last=False))
    check_repr(Convolution(n_filters=3, input_shape=(3, 4, 5), strides=(2, 2)))
    check_repr(Convolution(n_filters=3, input_shape=(3, 4, 5), padding="same"))
    assert (
        repr(Conv(n_filters=3, input_shape=(3, 4, 5)))
        == "Convolution(n_filters=3, input_shape=(3, 4, 5))"
    )
    assert (
        repr(Conv(n_filters=3, input_shape=(3, 4, 5), kernel_size=(3, 2)))
        == "Convolution(n_filters=3, input_shape=(3, 4, 5), kernel_size=(3, 2))"
    )
    assert (
        repr(Conv(n_filters=3, input_shape=(3, 4, 5), channels_last=False))
        == "Convolution(n_filters=3, input_shape=(3, 4, 5), channels_last=False)"
    )
    assert (
        repr(Conv(n_filters=3, input_shape=(3, 4, 5), strides=(2, 2)))
        == "Convolution(n_filters=3, input_shape=(3, 4, 5), strides=(2, 2))"
    )
    assert (
        repr(Conv(n_filters=3, input_shape=(3, 4, 5), padding="same"))
        == "Convolution(n_filters=3, input_shape=(3, 4, 5), padding='same')"
    )

    check_init_args(
        ConvTranspose,
        [
            "n_filters",
            "input_shape",
            "output_shape",
            "kernel_size",
            "strides",
            "padding",
            "channels_last",
            "init",
        ],
    )
    check_repr(ConvTranspose(n_filters=3, input_shape=(3, 4, 5)))
    check_repr(ConvTranspose(n_filters=3, input_shape=(3, 2, 5), kernel_size=(3, 2)))
    check_repr(ConvTranspose(n_filters=3, input_shape=(3, 4, 5), channels_last=False))
    check_repr(
        ConvolutionTranspose(
            n_filters=9, input_shape=(1, 2, 3), output_shape=(4, 5, 9), strides=(2, 2)
        )
    )
    assert (
        repr(ConvTranspose(n_filters=3, input_shape=(3, 4, 5)))
        == "ConvolutionTranspose(n_filters=3, input_shape=(3, 4, 5), "
        "output_shape=(5, 6, 3))"
    )
    assert (
        repr(ConvTranspose(n_filters=3, input_shape=(3, 4, 5), kernel_size=(3, 2)))
        == "ConvolutionTranspose(n_filters=3, input_shape=(3, 4, 5), "
        "output_shape=(5, 5, 3), kernel_size=(3, 2))"
    )
    assert (
        repr(ConvTranspose(n_filters=9, input_shape=(3, 4, 5), channels_last=False))
        == "ConvolutionTranspose(n_filters=9, input_shape=(3, 4, 5), "
        "output_shape=(9, 6, 7), channels_last=False)"
    )
    assert (
        repr(
            ConvolutionTranspose(
                n_filters=9,
                input_shape=(2, 3, 4),
                output_shape=(6, 8, 9),
                strides=(2, 2),
            )
        )
        == "ConvolutionTranspose(n_filters=9, input_shape=(2, 3, 4), "
        "output_shape=(6, 8, 9), strides=(2, 2))"
    )

    check_init_args(Sparse, ["shape", "indices", "init"])
    # No check_repr because sparse matrices are usually too big
    assert repr(Sparse((1, 1), indices=[[1, 1], [1, 1]])) == "Sparse(shape=(1, 1))"
    assert (
        repr(Sparse((1, 1), indices=[[1, 1], [1, 1], [1, 1]], init=2))
        == "Sparse(shape=(1, 1))"
    )

    check_init_args(SparseMatrix, ["indices", "data", "shape"])
    check_repr(SparseMatrix(indices=[[1, 2], [3, 4]], data=[5, 6], shape=(7, 8)))
    assert repr(SparseMatrix(((1, 2), (3, 4)), (5, 6), (7, 8))).replace(
        ", dtype=int64", ""
    ) == (
        "SparseMatrix(indices=array([[1, 2],\n       [3, 4]]), "
        "data=array([5, 6]), shape=(7, 8))"
    )

    check_init_args(ChannelShape, ["shape", "channels_last"])
    check_repr(ChannelShape(shape=(1, 2, 3), channels_last=True))
    assert (
        repr(ChannelShape((1, 2, 3)))
        == "ChannelShape(shape=(1, 2, 3), channels_last=True)"
    )
    assert (
        repr(ChannelShape((1, 2, 3), channels_last=False))
        == "ChannelShape(shape=(1, 2, 3), channels_last=False)"
    )

    # __str__ always has channels last
    assert str(ChannelShape((1, 2, 3))) == "(1, 2, ch=3)"
    assert str(ChannelShape((1, 2, 3), channels_last=False)) == "(ch=1, 2, 3)"

    check_init_args(NoTransform, ["size_in"])
    check_repr(NoTransform(size_in=1))
    for dimensions in range(2):
        assert repr(NoTransform(dimensions)) == f"NoTransform(size_in={dimensions})"


def test_signals():
    assert repr(Signal(np.array([0.0]))) == "Signal(name=None, shape=(1,))"
    assert (
        repr(Signal(np.array([1.0, 1.0]), name="one")) == "Signal(name=one, shape=(2,))"
    )


def test_operators():
    sig = Signal(np.array([0.0]), name="sig")
    assert fnmatch(repr(TimeUpdate(sig, sig)), "<TimeUpdate at 0x*>")
    assert fnmatch(repr(TimeUpdate(sig, sig, tag="tag")), "<TimeUpdate 'tag' at 0x*>")
    assert fnmatch(repr(Reset(sig)), "<Reset at 0x*>")
    assert fnmatch(repr(Reset(sig, tag="tag")), "<Reset 'tag' at 0x*>")
    assert fnmatch(repr(Copy(sig, sig)), "<Copy at 0x*>")
    assert fnmatch(repr(Copy(sig, sig, tag="tag")), "<Copy 'tag' at 0x*>")
    assert fnmatch(repr(ElementwiseInc(sig, sig, sig)), "<ElementwiseInc at 0x*>")
    assert fnmatch(
        repr(ElementwiseInc(sig, sig, sig, tag="tag")), "<ElementwiseInc 'tag' at 0x*>"
    )
    assert fnmatch(repr(DotInc(sig, sig, sig)), "<DotInc at 0x*>")
    assert fnmatch(repr(DotInc(sig, sig, sig, tag="tag")), "<DotInc 'tag' at 0x*>")
    assert fnmatch(repr(SimPyFunc(sig, lambda x: 0.0, True, sig)), "<SimPyFunc at 0x*>")
    assert fnmatch(
        repr(SimPyFunc(sig, lambda x: 0.0, True, sig, tag="tag")),
        "<SimPyFunc 'tag' at 0x*>",
    )
    assert fnmatch(repr(SimPES(sig, sig, sig, 0.1)), "<SimPES at 0x*>")
    assert fnmatch(repr(SimPES(sig, sig, sig, 0.1, tag="tag")), "<SimPES 'tag' at 0x*>")
    assert fnmatch(repr(SimBCM(sig, sig, sig, sig, 0.1)), "<SimBCM at 0x*>")
    assert fnmatch(
        repr(SimBCM(sig, sig, sig, sig, 0.1, tag="tag")), "<SimBCM 'tag' at 0x*>"
    )
    assert fnmatch(repr(SimOja(sig, sig, sig, sig, 0.1, 1.0)), "<SimOja at 0x*>")
    assert fnmatch(
        repr(SimOja(sig, sig, sig, sig, 0.1, 1.0, tag="tag")), "<SimOja 'tag' at 0x*>"
    )
    assert fnmatch(repr(SimVoja(sig, sig, sig, sig, 1.0, sig, 1.0)), "<SimVoja at 0x*>")
    assert fnmatch(
        repr(SimVoja(sig, sig, sig, sig, 0.1, sig, 1.0, tag="tag")),
        "<SimVoja 'tag' at 0x*>",
    )
    assert fnmatch(repr(SimRLS(sig, sig, sig, sig)), "<SimRLS at 0x*>")
    assert fnmatch(
        repr(SimRLS(sig, sig, sig, sig, tag="tag")),
        "<SimRLS 'tag' at 0x*>",
    )
    assert fnmatch(repr(SimNeurons(LIF(), sig, {"sig": sig})), "<SimNeurons at 0x*>")
    assert fnmatch(
        repr(SimNeurons(LIF(), sig, {"sig": sig}, tag="tag")),
        "<SimNeurons 'tag' at 0x*>",
    )
    assert fnmatch(repr(SimProcess(WhiteNoise(), sig, sig, sig)), "<SimProcess at 0x*>")
    assert fnmatch(
        repr(SimProcess(WhiteNoise(), sig, sig, sig, tag="tag")),
        "<SimProcess 'tag' at 0x*>",
    )



from nengo.utils.cache import byte_align, bytes2human, human2bytes


def test_bytes2human():
    assert bytes2human(1) == "1.0 B"
    assert bytes2human(10000) == "9.8 KB"
    assert bytes2human(100001221) == "95.4 MB"


def test_human2bytes():
    assert human2bytes("1 MB") == 1048576
    assert human2bytes("1.5 GB") == 1610612736
    assert human2bytes("14 B") == 14
    assert human2bytes("1B") == 1
    assert human2bytes("1   B") == 1


def test_byte_align():
    assert byte_align(5, 16) == 16
    assert byte_align(23, 8) == 24
    assert byte_align(13, 1) == 13
    assert byte_align(0, 16) == 0
    assert byte_align(32, 8) == 32



from .actions import Actions
from .assoc_mem import AssociativeMemory
from .basalganglia import BasalGanglia
from .bind import Bind
from .buffer import Buffer
from .compare import Compare
from .cortical import Cortical
from .input import Input
from .memory import Memory
from .pointer import SemanticPointer
from .spa import SPA
from .state import State
from .thalamus import Thalamus
from .utils import enable_spa_params, similarity
from .vocab import Vocabulary
